import os
from pathlib import Path
from functools import partial
import xml.etree.ElementTree as ET
import math
import sys
import json 
import logging      
import types                                                                                                                                                            

try:
    import numpy as np
    from fastai.vision.data import imagenet_stats, ImageList, bb_pad_collate
    from fastai.vision.transform import crop, rotate, dihedral_affine, brightness, contrast, skew, rand_zoom, get_transforms, flip_lr
    from fastai.vision import ImageDataBunch
    from fastai.torch_core import data_collate
    import torch
    from .models._ssd_utils import SSDObjectItemList
    from .models._unet_utils import ArcGISSegmentationItemList, ArcGISSegmentationMSItemList, _show_batch_unet_multispectral
    from .models._maskrcnn_utils import ArcGISInstanceSegmentationItemList
    from .models._ner_utils import ner_prepare_data
    HAS_FASTAI = True
except:
    HAS_FASTAI = False

band_abrevation_lib = {
    'b': 'BLUE',
    'c': 'CIRRUS',
    'ca': 'COASTAL AEROSOL',
    'g': 'GREEN',
    'nir': 'NEAR INFRARED',
    'nnir': 'NARROW NEAR INFRARED',
    'p': 'PANCHROMATIC',
    'r': 'RED',
    'swir': 'SHORT WAVELENGTH INFRARED',
    'swirc': 'SHORT WAVELENGTH INFRARED – Cirrus',
    'tir': 'THERMAL INFRARED',
    'vre': 'Vegetation red edge',
    'wv': 'WATER VAPOUR'
}

imagery_type_lib = {
    'landsat8': {
        "bands": ['ca', 'b', 'g', 'r', 'nir', 'swir', 'swir', 'c', 'qa', 'tir', 'tir'],
        "bands_info": { # incomplete
        }
    },
    "naip": {
        "bands": ['r', 'g', 'b', 'nir'],
        "bands_info": { # incomplete
        }
    },
    'sentinel2': { 
        "bands": ['ca', 'b', 'g', 'r', 'vre', 'vre', 'vre', 'nir', 'nnir', 'wv', 'swirc', 'swir', 'swir'],
        "bands_info": { # incomplete
            "b1": {
                "Name": "costal",
                "max": 10000,
                "min": 10000
            },            
            "b2": {
                "Name": "blue",
                "max": 10000,
                "min": 10000
            }
        }
    }
}

def _raise_fastai_import_error():
    raise Exception("""This module requires fastai, PyTorch, torchvision and scikit-image as its dependencies. 
Install them using 'conda install -c pytorch -c fastai fastai=1.0.54 pytorch=1.1.0 torchvision scikit-image'""")

def _bb_pad_collate(samples, pad_idx=0):
    "Function that collect `samples` of labelled bboxes and adds padding with `pad_idx`."
    if isinstance(samples[0][1], int):
        return data_collate(samples)
    max_len = max([len(s[1].data[1]) for s in samples])
    bboxes = torch.zeros(len(samples), max_len, 4)
    labels = torch.zeros(len(samples), max_len).long() + pad_idx
    imgs = []
    for i,s in enumerate(samples):
        imgs.append(s[0].data[None])
        bbs, lbls = s[1].data

        if not (bbs.nelement() == 0) or list(bbs) == [[0,0,0,0]]:
            bboxes[i,-len(lbls):] = bbs
            labels[i,-len(lbls):] = torch.tensor(lbls, device=bbs.device).long()
    return torch.cat(imgs,0), (bboxes,labels)    


def _get_bbox_classes(xmlfile, class_mapping, not_label_count=[0], height_width=[]):
    if not os.path.exists(xmlfile):
        not_label_count[0] += 1
        return [[[0, 0, 0, 0]], [list(class_mapping.values())[0]]]

    tree = ET.parse(xmlfile)
    xmlroot = tree.getroot()
    bboxes = []
    classes = []
    for tag_obj in xmlroot.findall('object'):
        bnd_box = tag_obj.find('bndbox')
        xmin, ymin, xmax, ymax = float(bnd_box.find('xmin').text), \
                                 float(bnd_box.find('ymin').text), \
                                 float(bnd_box.find('xmax').text), \
                                 float(bnd_box.find('ymax').text)
        data_class_text = tag_obj.find('name').text
        
        if (not data_class_text.isnumeric() and not class_mapping.get(data_class_text))\
             or (data_class_text.isnumeric() and not(class_mapping.get(data_class_text) or class_mapping.get(int(data_class_text)))):
            continue

        if data_class_text.isnumeric():
            data_class_mapping = class_mapping[data_class_text] if class_mapping.get(data_class_text) else class_mapping[int(data_class_text)]
        else:
            data_class_mapping = class_mapping[data_class_text]

        classes.append(data_class_mapping)
        bboxes.append([ymin, xmin, ymax, xmax])
        height_width.append(((xmax - xmin)*1.25, (ymax - ymin)*1.25))
    
    if len(bboxes) == 0:
        return [[[0, 0, 0, 0]], [list(class_mapping.values())[0]]]
    return [bboxes, classes]


def _get_bbox_lbls(imagefile, class_mapping, not_label_count, height_width):
    xmlfile = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), '.xml')
    return _get_bbox_classes(xmlfile, class_mapping, not_label_count, height_width)


def _get_lbls(imagefile, class_mapping):
    xmlfile = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), '.xml')
    return _get_bbox_classes(xmlfile, class_mapping)[1][0]

def _check_esri_files(path):
    if os.path.exists(path / 'esri_model_definition.emd') \
        and os.path.exists(path / 'map.txt') \
            and os.path.exists(path / 'esri_accumulated_stats.json'):
        return True

    return False

def _get_class_mapping(path):
    class_mapping = {}
    for xmlfile in os.listdir(path):
        if not xmlfile.endswith('.xml'):
            continue
        tree = ET.parse(os.path.join(path, xmlfile))
        xmlroot = tree.getroot()
        for tag_obj in xmlroot.findall('object'):
            class_mapping[tag_obj.find('name').text] = tag_obj.find('name').text

    return class_mapping

def _get_batch_stats(image_list, norm_pct=1, _band_std_values=False):
    n_normalization_samples = round(len(image_list)*norm_pct)
    #n_normalization_samples = max(256, n_normalization_samples)
    random_indexes = np.random.randint(0, len(image_list), size=min(n_normalization_samples, len(image_list)))

    # Original Band Stats
    min_values_store = []
    max_values_store = []
    mean_values_store = []

    data_shape = image_list[0].data.shape
    n_bands = data_shape[0]
    feasible_chunk = round(512*4*400/(n_bands*data_shape[1])) # ~3gb footprint
    chunk = min(feasible_chunk, n_normalization_samples)
    i = 0
    for i in range(0, n_normalization_samples, chunk):
        x_tensor_chunk = torch.stack([ x.data for x in image_list[random_indexes[i:i+chunk]] ] )
        """
        min_values = torch.zeros(n_bands)
        max_values = torch.zeros(n_bands)
        mean_values = torch.zeros(n_bands)
        for bi in range(n_bands):
            min_values[bi] = x_tensor_chunk[:, bi].min()
            max_values[bi] = x_tensor_chunk[:, bi].max()
            mean_values[bi] = x_tensor_chunk[:, bi].mean()
        """
        min_values = x_tensor_chunk.min(dim=0)[0].min(dim=1)[0].min(dim=1)[0]
        max_values = x_tensor_chunk.max(dim=0)[0].max(dim=1)[0].max(dim=1)[0]
        mean_values = x_tensor_chunk.mean((0, 2, 3))
        min_values_store.append(min_values)
        max_values_store.append(max_values)
        mean_values_store.append(mean_values)

    band_max_values = torch.stack(max_values_store).max(dim=0)[0]
    band_min_values = torch.stack(min_values_store).min(dim=0)[0]
    band_mean_values = torch.stack(mean_values_store).mean(dim=0)
    
    view_shape = _get_view_shape(image_list[0].data, band_mean_values)

    if _band_std_values:
        std_values_store = []
        for i in range(0, n_normalization_samples, chunk):
            x_tensor_chunk = torch.stack([ x.data for x in image_list[random_indexes[i:i+chunk]] ] )
            std_values = (x_tensor_chunk - band_mean_values.view(view_shape)).pow(2).sum((0, 2, 3))
            std_values_store.append(std_values)
        band_std_values = (torch.stack(std_values_store).sum(dim=0) / ((n_normalization_samples * data_shape[1] * data_shape[2])-1)).sqrt()
    else:
        band_std_values = None

    # Scaled Stats
    scaled_min_values = torch.tensor([0 for i in range(n_bands)], dtype=torch.float32)
    scaled_max_values = torch.tensor([1 for i in range(n_bands)], dtype=torch.float32)
    scaled_mean_values = _tensor_scaler(band_mean_values, band_min_values, band_max_values, mode='minmax')
    
    scaled_std_values_store = []
    for i in range(0, n_normalization_samples, chunk):
        x_tensor_chunk = torch.stack([ x.data for x in image_list[random_indexes[i:i+chunk]] ] )
        x_tensor_chunk = _tensor_scaler(x_tensor_chunk, band_min_values, band_max_values, mode='minmax')
        std_values = (x_tensor_chunk - scaled_mean_values.view(view_shape)).pow(2).sum((0, 2, 3))
        scaled_std_values_store.append(std_values)
    scaled_std_values = (torch.stack(scaled_std_values_store).sum(dim=0) / ((n_normalization_samples * data_shape[1] * data_shape[2])-1)).sqrt()

    #return band_min_values, band_max_values, band_mean_values, band_std_values, scaled_min_values, scaled_max_values, scaled_mean_values, scaled_std_values
    return {
        "band_min_values": band_min_values, 
        "band_max_values": band_max_values, 
        "band_mean_values": band_mean_values, 
        "band_std_values": band_std_values, 
        "scaled_min_values": scaled_min_values, 
        "scaled_max_values": scaled_max_values, 
        "scaled_mean_values": scaled_mean_values, 
        "scaled_std_values": scaled_std_values
    }


def _get_view_shape(tensor_batch, band_factors):
    view_shape = [1 for i in range(len(tensor_batch.shape))]
    view_shape[tensor_batch.shape.index(band_factors.shape[0])] = band_factors.shape[0]
    return tuple(view_shape)

def _tensor_scaler(tensor_batch, min_values, max_values, mode='minmax', create_view=True):
    if create_view:
        view_shape = _get_view_shape(tensor_batch, min_values)
        max_values = max_values.view(view_shape)
        min_values = min_values.view(view_shape)
    if mode=='minmax':
        # new_value = (((old_value - old_min) / (old_max-old_min))*(new_max-new_min))+new_min
        scaled_tensor_batch = (tensor_batch - min_values) / (max_values - min_values)
    return scaled_tensor_batch

def _tensor_scaler_tfm(tensor_batch, min_values, max_values, mode='minmax'):
    x = tensor_batch[0]
    y = tensor_batch[1]
    max_values = max_values.view(1, -1, 1, 1).to(x.device)
    min_values = min_values.view(1, -1, 1, 1).to(x.device)
    x = _tensor_scaler(x, min_values, max_values, mode, create_view=False)
    return (x, y)

def _extract_bands_tfm(tensor_batch, band_indices):
    x_batch = tensor_batch[0][:, band_indices]
    y_batch = tensor_batch[1]
    return (x_batch, y_batch)

def prepare_data(path,
                 class_mapping=None, 
                 chip_size=224, 
                 val_split_pct=0.1, 
                 batch_size=64, 
                 transforms=None, 
                 collate_fn=_bb_pad_collate, 
                 seed=42, 
                 dataset_type=None, 
                 resize_to=None,
                 **kwargs):
    """
    Prepares a data object from training sample exported by the 
    Export Training Data tool in ArcGIS Pro or Image Server, or training 
    samples in the supported dataset formats. This data object consists of 
    training and validation data sets with the specified transformations, 
    chip size, batch size, split percentage, etc. 
    -For object detection, use Pascal_VOC_rectangles format.
    -For feature categorization use Labelled Tiles or ImageNet format.
    -For pixel classification, use Classified Tiles format.
    -For entity extraction from text, use BIO, LBIOU or ner_json formats. 

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    path                    Required string. Path to data directory.
    ---------------------   -------------------------------------------
    class_mapping           Optional dictionary. Mapping from id to
                            its string label.
                            For dataset_type=BIO, LBIOU or ner_json:
                                Provide address field as class mapping
                                in below format:
                                class_mapping={'address_tag':'address_field'}
    ---------------------   -------------------------------------------
    chip_size               Optional integer. Size of the image to train the
                            model.
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors).
    ---------------------   -------------------------------------------
    transforms              Optional tuple. Fast.ai transforms for data
                            augmentation of training and validation datasets
                            respectively (We have set good defaults which work
                            for satellite imagery well). If transforms is set
                            to `False` no transformation will take place and 
                            `chip_size` parameter will also not take effect.
    ---------------------   -------------------------------------------
    collate_fn              Optional function. Passed to PyTorch to collate data
                            into batches(usually default works).
    ---------------------   -------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
    ---------------------   -------------------------------------------
    dataset_type            Optional string. `prepare_data` function will infer 
                            the `dataset_type` on its own if it contains a 
                            map.txt file. If the path does not contain the 
                            map.txt file pass either of 'PASCAL_VOC_rectangles', 
                            'RCNN_Masks' and 'Classified_Tiles'                    
    ---------------------   -------------------------------------------
    resize_to               Optional integer. Resize the image to given size.
    =====================   ===========================================

    :returns: data object
    """
    """kwargs documentation
    imagery_type='RGB'
    bands=None
    rgb_bands=[0, 1, 2]
    """

    height_width = []

    if not HAS_FASTAI:
        _raise_fastai_import_error()

    if type(path) is str:
        path = Path(path)

    databunch_kwargs = {'num_workers':0} if sys.platform == 'win32' else {}
    databunch_kwargs['bs'] = batch_size

    kwargs_transforms = {}
    if resize_to:
        kwargs_transforms['size'] = resize_to

    has_esri_files = _check_esri_files(path)
    alter_class_mapping = False
    color_mapping = None

    # Multispectral Kwargs init
    _bands = None
    _imagery_type = None
    _is_multispectral = False
    _show_batch_multispectral = None

    if dataset_type is None and not has_esri_files:
        raise Exception("Could not infer dataset type.")

    if dataset_type != "Imagenet" and has_esri_files:
        stats_file = path / 'esri_accumulated_stats.json'
        with open(stats_file) as f:
            stats = json.load(f)
            dataset_type = stats['MetaDataMode']

        with open(path / 'map.txt') as f:
            line = f.readline()

        right = line.split()[1].split('.')[-1].lower()

        json_file = path / 'esri_model_definition.emd'
        with open(json_file) as f:
            emd = json.load(f)

        # Create Class Mapping from EMD if not specified by user
        if class_mapping is None:
            try:
                class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
            except KeyError:
                class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}

        color_mapping = {(i.get('Value', 0) or i.get('ClassValue', 0)): i['Color'] for i in emd.get('Classes', [])}

        if color_mapping.get(None):
            del color_mapping[None]

        if class_mapping.get(None):
            del class_mapping[None]

        # Multispectral support from EMD 
        # Not Implemented Yet
        if emd.get('bands', None) is not None: 
            _bands = emd.get['bands'] # Not Implemented
            
        if emd.get('imagery_type', None) is not None: 
            _imagery_type = emd.get['imagery_type'] # Not Implemented        

    elif dataset_type == 'PASCAL_VOC_rectangles' and not has_esri_files:
        if class_mapping is None:
            class_mapping = _get_class_mapping(path / 'labels')
            alter_class_mapping = True

    # Multispectral check
    imagery_type = 'RGB'
    if kwargs.get('imagery_type', None) is not None:
        imagery_type = kwargs.get('imagery_type')
    elif _imagery_type is not None:
        imagery_type = _imagery_type

    bands = None
    if kwargs.get('bands', None) is not None:
        bands = kwargs.get('bands')
        for i, b in enumerate(bands):
            if type(b) == str:
                bands[i] = b.lower()
    elif imagery_type_lib.get(imagery_type, None) is not None:
        bands = imagery_type_lib.get(imagery_type)['bands']
    elif _bands is not None:
        bands = _bands

    rgb_bands = None
    if kwargs.get('rgb_bands', None) is not None:
        rgb_bands = kwargs.get('rgb_bands')
    elif bands is not None:
        rgb_bands = [ bands.index(b) for b in ['r', 'g', 'b'] if b in bands ]
    
    if (bands is not None) or (rgb_bands is not None) or (not imagery_type == 'RGB'):
        if imagery_type == 'RGB':
            imagery_type = 'multispectral'
        _is_multispectral = True
    
    if kwargs.get('norm_pct', None) is not None:
        norm_pct= kwargs.get('norm_pct')
        norm_pct = min(max(0, norm_pct), 1)
    else:
        norm_pct = .1

    if dataset_type == 'RCNN_Masks':

        def get_labels(x, label_dirs, ext=right):
            label_path = []
            for lbl in label_dirs:
                if os.path.exists(Path(lbl) / (x.stem + '.{}'.format(ext))):
                    label_path.append(Path(lbl) / (x.stem + '.{}'.format(ext)))
            return label_path

        if class_mapping.get(0):
            del class_mapping[0]

        if color_mapping.get(0):
            del color_mapping[0]

      
        src = (ArcGISInstanceSegmentationItemList.from_folder(path/'images')
            .split_by_rand_pct(val_split_pct, seed=seed))

        label_dirs = os.listdir(path/'labels')
        label_dir = [os.path.join(path/'labels', lbl) for lbl in label_dirs if os.path.isdir(os.path.join(path/'labels', lbl))]
        get_y_func = partial(get_labels, label_dirs= label_dir)
        src = src.label_from_func(get_y_func, chip_size=chip_size, classes=['NoData'] + list(class_mapping.values()), class_mapping=class_mapping, color_mapping=color_mapping)
    
    elif dataset_type == 'Classified_Tiles':

        def get_y_func(x, ext=right):
            return x.parents[1] / 'labels' / (x.stem + '.{}'.format(ext))

        if class_mapping.get(0):
            del class_mapping[0]

        if color_mapping.get(0):
            del color_mapping[0]

        # TODO : Handle NoData case

        # Handle Multispectral
        if _is_multispectral:
            data = ArcGISSegmentationMSItemList.from_folder(path/'images')\
                .split_by_rand_pct(val_split_pct, seed=seed)\
                .label_from_func(
                    get_y_func, classes=(['NoData'] + list(class_mapping.values())),
                    class_mapping=class_mapping,
                    color_mapping=color_mapping
                )
            _show_batch_multispectral = _show_batch_unet_multispectral            

            def classified_tiles_collate_fn(samples): # The default fastai collate_fn was causing memory leak on tensors
                r = ( torch.stack([x[0].data for x in samples]), torch.stack([x[1].data for x in samples]) )
                return r
            databunch_kwargs['collate_fn'] = classified_tiles_collate_fn

        else:
            data = ArcGISSegmentationItemList.from_folder(path/'images')\
                .split_by_rand_pct(val_split_pct, seed=seed)\
                .label_from_func(
                    get_y_func, classes=(['NoData'] + list(class_mapping.values())),
                    class_mapping=class_mapping,
                    color_mapping=color_mapping
                )

        if transforms is None:
            transforms = get_transforms(
                flip_vert=True,
                max_rotate=90.,
                max_zoom=3.0,
                max_lighting=0.5
            )

        kwargs_transforms['tfm_y'] = True
        kwargs_transforms['size'] = chip_size
    elif dataset_type == 'PASCAL_VOC_rectangles':
        not_label_count = [0]
        get_y_func = partial(
            _get_bbox_lbls,
            class_mapping=class_mapping,
            not_label_count=not_label_count,
            height_width=height_width
        )

        data = SSDObjectItemList.from_folder(path/'images')\
            .split_by_rand_pct(val_split_pct, seed=seed)\
            .label_from_func(get_y_func)

        if not_label_count[0]:
            logger = logging.getLogger()
            logger.warning("Please check your dataset. " + str(not_label_count[0]) + " images dont have the corresponding label files.")

        if transforms is None:
            ranges = (0, 1)
            train_tfms = [
                crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                dihedral_affine() if has_esri_files else flip_lr(),
                brightness(change=(0.4, 0.6)),
                contrast(scale=(0.75, 1.5)),
                rand_zoom(scale=(1.0, 1.5))
            ]
            val_tfms = [crop(size=chip_size, p=1., row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)

        kwargs_transforms['tfm_y'] = True
        databunch_kwargs['collate_fn'] = collate_fn
    elif dataset_type in ['Labeled_Tiles', 'Imagenet']:
        if dataset_type == 'Labeled_Tiles':
            get_y_func = partial(_get_lbls, class_mapping=class_mapping)
        else:
            def get_y_func(x):
                return x.parent.stem

        data = ImageList.from_folder(path/'images')\
            .split_by_rand_pct(val_split_pct, seed=42)\
            .label_from_func(get_y_func)

        if dataset_type == 'Imagenet':
            class_mapping = {}
            index = 1
            for class_name in data.classes:
                class_mapping[index] = class_name
                index = index + 1

        if transforms is None:
            ranges = (0, 1)
            train_tfms = [
                rotate(degrees=30, p=0.5),
                crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                dihedral_affine(),
                brightness(change=(0.4, 0.6)),
                contrast(scale=(0.75, 1.5))
            ]
            val_tfms = [crop(size=chip_size, p=1.0, row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)
    elif dataset_type in ['ner_json','BIO','LBIOU']:
        return ner_prepare_data(dataset_type=dataset_type, path=path, class_mapping=class_mapping, val_split_pct=val_split_pct)
    else:
        raise NotImplementedError('Unknown dataset_type="{}".'.format(dataset_type))
    
    if dataset_type == 'RCNN_Masks':
        if transforms ==  None:
            data = (src.transform(size=chip_size, tfm_y=True)
                .databunch(**databunch_kwargs))
        else:
            data = (src.transform(transforms, size=chip_size, tfm_y=True) 
                    .databunch(**databunch_kwargs))
    elif _is_multispectral:
        data = data.databunch(**databunch_kwargs)

        # Statistics        
        dummy_stats = {
            "batch_stats_for_norm_pct_0" : {
                "band_min_values":None, 
                "band_max_values":None, 
                "band_mean_values":None, 
                "band_std_values":None, 
                "scaled_min_values":None, 
                "scaled_max_values":None, 
                "scaled_mean_values":None, 
                "scaled_std_values":None
            }
        }
        normstats_json_path = os.path.abspath(data.path / '..' / 'esri_normalization_stats.json')
        if not os.path.exists(normstats_json_path):
            normstats = dummy_stats
            with open(normstats_json_path, 'w', encoding='utf-8') as f:
                json.dump(normstats, f, ensure_ascii=False, indent=4)
        else:
            with open(normstats_json_path) as f:
                normstats = json.load(f)

        norm_pct_search = f"batch_stats_for_norm_pct_{round(norm_pct*100)}"
        if norm_pct_search in normstats:
            batch_stats = normstats[norm_pct_search]
            for s in batch_stats:
                if batch_stats[s] is not None:
                    batch_stats[s] = torch.tensor(batch_stats[s])
        else:
            batch_stats = _get_batch_stats(data.x, norm_pct)
            normstats[norm_pct_search] = dict(batch_stats)
            for s in normstats[norm_pct_search]:
                if normstats[norm_pct_search][s] is not None:
                    normstats[norm_pct_search][s] = normstats[norm_pct_search][s].tolist()
            with open(normstats_json_path, 'w', encoding='utf-8') as f:
                json.dump(normstats, f, ensure_ascii=False, indent=4)

        # batch_stats -> [band_min_values, band_max_values, band_mean_values, band_std_values, scaled_min_values, scaled_max_values, scaled_mean_values, scaled_std_values]
        data._band_min_values = batch_stats['band_min_values']
        data._band_max_values = batch_stats['band_max_values']
        data._band_mean_values = batch_stats['band_mean_values']
        data._band_std_values = batch_stats['band_std_values']
        data._scaled_min_values = batch_stats['scaled_min_values']
        data._scaled_max_values = batch_stats['scaled_max_values']
        data._scaled_mean_values = batch_stats['scaled_mean_values']
        data._scaled_std_values = batch_stats['scaled_std_values']
        
        # Scaling
        if kwargs.get('do_scale', None) is not None:
            data._do_scale = kwargs.get('do_scale')
        else:
            data._do_scale = True
        data._min_max_scaler = partial(_tensor_scaler, min_values=data._band_min_values, max_values=data._band_max_values, mode='minmax')
        if data._do_scale:
            data._min_max_scaler_tfm = partial(_tensor_scaler_tfm, min_values=data._band_min_values, max_values=data._band_max_values, mode='minmax')
            data.add_tfm(data._min_max_scaler_tfm)
        
        # Normalize
        if kwargs.get('do_normalize', None) is not None:
            data._do_normalize = kwargs.get('do_normalize')
        else:
            data._do_normalize = True
        if data._do_normalize:
            data = data.normalize(stats=(data._scaled_mean_values, data._scaled_std_values), do_x=True, do_y=False)
    else:
        # 
        data = (data.transform(transforms, **kwargs_transforms)
            .databunch(**databunch_kwargs)
            .normalize(imagenet_stats))


    data.chip_size = data.x[0].shape[-1] if transforms is False else chip_size

    if alter_class_mapping:
        new_mapping = {}
        for i, class_name in enumerate(class_mapping.keys()):
            new_mapping[i+1] = class_name
        class_mapping = new_mapping

    data.class_mapping = class_mapping
    data.color_mapping = color_mapping
    data.show_batch = partial(data.show_batch, rows=min(int(math.sqrt(batch_size)), 5))
    data.orig_path = path
    data.resize_to = kwargs_transforms.get('size', None)
    data.height_width = height_width
    
    data._is_multispectral = _is_multispectral
    if data._is_multispectral:
        data._imagery_type = imagery_type
        data._bands = bands
        data._norm_pct = norm_pct
        data._rgb_bands = rgb_bands
        data._symbology_rgb_bands = rgb_bands

        # Handle invalid color mapping
        data._multispectral_color_mapping = color_mapping
        if any( -1 in x for x in data._multispectral_color_mapping.values() ):
            random_color_list = np.random.randint(low=0, high=255, size=(len(data._multispectral_color_mapping), 3)).tolist()
            for i, c in enumerate(data._multispectral_color_mapping):
                if -1 in data._multispectral_color_mapping[c]:
                    data._multispectral_color_mapping[c] = random_color_list[i]

        # Prepare unknown bands list if bands data is missing
        if data._bands is None:
            n_bands = data.x[0].data.shape[0]
            if n_bands == 1:# Handle Pancromatic case
                data._bands = data._symbology_rgb_bands = data._rgb_bands = ['p']
            else:
                data._bands = ['u' for i in range(n_bands)]
        
        # 
        if data._rgb_bands is None:
            data._rgb_bands = []

        # 
        if data._symbology_rgb_bands is None:
            data._symbology_rgb_bands = [0, 1, 2][:min(n_bands, 3)]

        # Complete symbology rgb bands
        if len(data._bands) > 2 and len(data._symbology_rgb_bands) < 3:
            data._symbology_rgb_bands += [ min(max(data._symbology_rgb_bands)+1, len(data._bands)-1)  for i in range(3 - len(data._symbology_rgb_bands)) ]
        
        # Overwrite band values at r g b indexes with 'r' 'g' 'b'
        for i, band_idx in enumerate(data._rgb_bands):
            if band_idx is not None:
                if data._bands[band_idx] == 'u':
                    data._bands[band_idx] = ['r', 'g', 'b'][i]

        # Attach custom show batch
        if _show_batch_multispectral is not None:
            data.show_batch = types.MethodType( _show_batch_multispectral, data )

        # Apply filter band transformation if user has specified extract_bands otherwise add a generic extract_bands
        """
        extract_bands : List containing band indices of the bands from imagery on which the model would be trained. 
                        Useful for benchmarking and applied training, for reference see examples below.
                        
                        4 band naip ['r, 'g', 'b', 'nir'] + extract_bands=[0, 1, 2] -> 3 band naip with bands ['r', 'g', 'b'] 

        """
        data._extract_bands = kwargs.get('extract_bands', None) 
        if data._extract_bands is None:
            data._extract_bands = list(range(len(data._bands)))
        else:
            data._extract_bands_tfm = partial(_extract_bands_tfm, band_indices=data._extract_bands)
            data.add_tfm(data._extract_bands_tfm)   

        # Tail Training Override
        _train_tail = True
        if [data._bands[i] for i in data._extract_bands] == ['r', 'g', 'b']:
            _train_tail = False
        data._train_tail = kwargs.get('train_tail', _train_tail)

    return data



try:
    from arcgis.widgets._mapview import MapView
except ImportError as e:
    import logging
    log = logging.getLogger()
    import_error = e
    class MapView:
        def __init__(self, *args, **kwargs):
            log.warning("MapView class replaced with a non-functional "\
                        "placeholder due to the following import error:")
            raise import_error




class LocationTrackingError(Exception):
    """
    Location Tracking Error Class
    """
    def __init__(self, message):
        super().__init__(self)
        self.message = message

    def __str__(self):
        return self.message



"""
The Summarize Data module contains functions that calculate total counts, lengths, areas, and basic descriptive statistics of features and their attributes within areas or near other features.

aggregate_points calculates statistics about points that fall within specified areas or bins.
join_features calculates statistics about features that share a spatial, temporal, or attribute relationship with other features.
reconstruct_tracks calculates statistics about points or polygons that belong to the same track and reconstructs inputs into tracks.
summarize_attributes calculates statistics about feature or tabular data that share attributes.
summarize_within calculates statistics for area features and attributes that overlap each other.
"""

import json as _json
import datetime
from datetime import datetime as _datetime
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.features import Table as _Table
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoprocessing import DataFile
from ._util import _id_generator, _feature_input, _set_context, _create_output_service, GAJob

_log = _logging.getLogger(__name__)

# url = "https://dev003153.esri.com/gax/rest/services/System/GeoAnalyticsTools/GPServer"

_use_async = True

def build_multivariable_grid(input_layers,
                             variable_calculations,
                             bin_size,
                             bin_unit="Meters",
                             bin_type="Square",
                             output_name=None,
                             gis=None,
                             future=False):
    """

    .. image:: _static/images/Grid/Grid.png

    The ``build_multivariable_grid`` task works with one or more layers of point, line, or polygon features.
    The task generates a grid of square or hexagonal bins and compiles information about each input layer into each bin.
    For each input layer, this information can include the following variables:

        * ``Distance to Nearest`` - The distance from each bin to the nearest feature.
        * ``Attribute of Nearest`` - An attribute value of the feature nearest to each bin.
        * ``Attribute Summary of Related`` - A statistical summary of all features within ``search_distance`` of each bin.

    Only variables you specify in ``variable_calculations`` will be included in the result layer. These variables can help
    you understand the proximity of your data throughout the extent of your analysis. The results can help you answer
    questions such as the following:

        * Given multiple layers of public transportation infrastructure, what part of the city is least accessible by public transportation?
        * Given layers of lakes and rivers, what is the name of the water body closest to each location in the U.S.?
        * Given a layer of household income, where in the U.S. is the variation of income in the surrounding 50 miles the greatest?

    The result of ``build_multivariable_grid`` can also be used in prediction and classification workflows. The task allows you
    to calculate and compile information from many different data sources into a single, spatially continuous layer in one step.
    This layer can then be used with the Enrich From Multi-Variable Grid task to quickly enrich point features with the variables
    you have calculated, reducing the amount of effort required to build prediction and classification models from point data.

    ===================================================================    =============================================================================
    **Argument**                                                                                    **Description**
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    input_layers                                                           Required list of layers. A list of input layers that will be used in analysis.
                                                                           See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    variable_calculations                                                  Required list of dicts. A dict containing objects that describe
                                                                           the variables that will be calculated for each layer in ``input_layers``.

                                                                           [
                                                                                {
                                                                                    "layer":<index>,
                                                                                    "variables":[
                                                                                        {
                                                                                            "type":"DistanceToNearest",
                                                                                            "outFieldName":"<output field name>",
                                                                                            "searchDistance":<number>,
                                                                                            "searchDistanceUnit":"<unit>",
                                                                                            "filter":"<filter>"
                                                                                        },
                                                                                        {
                                                                                            "type":"AttributeOfNearest",
                                                                                            "outFieldName":"<output field name>",
                                                                                            "attributeField":"<field name>",
                                                                                            "searchDistance":<number>,
                                                                                            "searchDistanceUnit":"<unit>",
                                                                                            "filter":"<filter>"
                                                                                        },
                                                                                        {
                                                                                            "type":"AttributeSummaryOfRelated",
                                                                                            "outFieldName":"<output field name>",
                                                                                            "statisticType":"<statistic type>",
                                                                                            "statisticField":"<field name>",
                                                                                            "searchDistance":<number>,
                                                                                            "searchDistanceUnit":"<unit>",
                                                                                            "filter":"<filter>"
                                                                                        },
                                                                                        ...
                                                                                    ]
                                                                                },
                                                                                ...
                                                                            ]

                                                                           layer is the index of the layer in ``input_layers`` that will be
                                                                           used to calculate the specified variables.

                                                                           Variables is an array of dict objects that describe the variables
                                                                           you want to include in the result layer. The array must contain at least
                                                                           one variable for each layer.

                                                                           type can be one of the following variable types:

                                                                                * DistanceToNearest
                                                                                * AttributeOfNearest
                                                                                * AttributeSummaryOfRelated

                                                                           Each type must be configured with a unique set of parameters:

                                                                            * ``outFieldName`` is the name of the field that will be created in the result
                                                                              layer to store a variable. This is required.
                                                                            * ``searchDistance`` and searchDistanceUnit are a number and linear unit.
                                                                              For DistanceToNearest and AttributeOfNearest, searchDistance and searchDistanceUnit
                                                                              are required and define the maximum distance that the tool will search from the
                                                                              center of each bin to find a feature in the layer. If no feature is within the
                                                                              distance, null is returned. For AttributeSummaryOfRelated, searchDistance and
                                                                              searchDistanceUnit are optional and define the radius of a circular neighborhood
                                                                              surrounding each bin. All features that intersect this neighborhood will be used
                                                                              to calculate statisticType. If a distance is not defined, only features that
                                                                              intersect a bin will be used to calculate statisticType.
                                                                            * ``attributeField`` is required by AttributeOfNearest and is the name of a field `
                                                                              in the input layer. The value of this field in the closest feature to each bin will
                                                                              be included in the result layer.
                                                                            * ``statisticField`` is required by AttributeSummaryOfRelated and is the name of a
                                                                              field in the input layer. This field's values will be used to calculate statisticType.
                                                                            * ``statisticType`` is required by AttributeSummaryOfRelated and is one of the following
                                                                              when statisticField is a numeric field:

                                                                                * ``Count`` - Totals the number of features near or intersecting each bin.
                                                                                * ``Sum`` - Adds the total value of all features near or intersecting each bin.
                                                                                * ``Mean`` - Calculates the average of all features near or intersecting each bin.
                                                                                * ``Min`` - Finds the smallest value of all features near or intersecting each bin.
                                                                                * ``Max`` - Finds the largest value of all features near or intersecting each bin.
                                                                                * ``Range`` - Finds the difference between Min and Max.
                                                                                * ``Stddev`` - Finds the standard deviation of all features near or intersecting each bin.
                                                                                * ``Var`` - Finds the variance of all features near or intersecting each bin.

                                                                            * ``statisticType`` is one of the following when statisticField is a string field:

                                                                                * Count Totals the number of strings for all features near or intersecting each bin.
                                                                                * Any Returns a sample string of all features near or intersecting each bin.

                                                                            * ``filter`` is optional for all variable types and is formatted as described in the Feature Input topic.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    bin_size                                                               Required float. The distance for the bins of type ``bin_type`` in the output polygon layer.
                                                                           ``variable_calculations`` will be calculated at the center of each bin. When generating bins,
                                                                           for Square, the number and units specified determine the height and length of the square.
                                                                           For Hexagon, the number and units specified determine the distance between parallel sides.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    bin_unit                                                               Optional string. The distance unit for the bins that will be used to calculate ``variable_calculations``.

                                                                           Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    bin_type                                                               Optional string. The type of bin that will be used to generate the result grid. Bin options are the following:

                                                                           Choice list: ['Hexagon', 'Square']

                                                                           .. Note::
                                                                            Analysis using Square or Hexagon bins requires a projected coordinate system.
                                                                            When aggregating layers into bins, the input layers or processing extent (``processSR``)
                                                                            must have a projected coordinate system. If a projected coordinate system is not
                                                                            specified when running analysis, the World Cylindrical Equal Area (WKID 54034) projection
                                                                            will be used. At 10.7 or later, if a projected coordinate system is not specified when
                                                                            running analysis, a projection will be picked based on the extent of the data.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    output_name                                                            Optional string. The task will create a feature service of the results. You define the name of the service.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    context                                                                Optional string. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                                           #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                           #. Processing spatial reference (``processSR``) The features will be projected into this coordinate system for analysis.
                                                                           #. Output Spatial Reference (``outSR``) - the features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                           #. Data store (``dataStore``) Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ===================================================================    =============================================================================

    :returns: boolean

    .. code-block:: python

            # Usage Example: To create multivariable grid by summarizing information such as distance to nearest

            variables = [ { "layer":0,
                            "variables":[
                                { "type":"DistanceToNearest",
                                  "outFieldName":"road",
                                  "searchDistance":10,
                                  "searchDistanceUnit":"Kilometers"
                                }
                            ]
                          },
                          { "layer":1,
                          "variables":[
                              { "type":"AttributeSummaryOfRelated",
                                "outFieldName":"MeanPopAge",
                                "statisticType":"Mean",
                                "statisticField":"Age",
                                "searchDistance":50,
                                "searchDistanceUnit":"Kilometers"
                              }
                          ]
                          }
                        ]
            grid = build_multivariable_grid(input_layers=[lyr0, lyr1],
                                            variable_calculations=variables,
                                            bin_size=100,
                                            bin_unit='Meters',
                                            bin_type='Square',
                                            output_name="multi_variable_grid")
    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if key == 'variable_calculations':
            import json
            params[key] = json.dumps(value)
        elif value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Build Multi Variable Grid_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Build Multi Variable Grid ')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db={
        "input_layers": (_FeatureSet, "inputLayers"),
        "variable_calculations" : (str, "variableCalculations"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_unit": (str, "binSizeUnit"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:

            gpjob = _execute_gp_tool(gis, "BuildMultiVariableGrid", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "BuildMultiVariableGrid", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise


def aggregate_points(point_layer,
                     bin_type=None,
                     bin_size=None,
                     bin_size_unit=None,
                     polygon_layer=None,
                     time_step_interval=None,
                     time_step_interval_unit=None,
                     time_step_repeat_interval=None,
                     time_step_repeat_interval_unit=None,
                     time_step_reference=None,
                     summary_fields=None,
                     output_name=None,
                     gis=None,
                     future=False):
    """
    .. image:: _static/images/aggregate_points/aggregate_points.png

    This ``aggregate_points`` tool works with a layer of point features and a layer of areas.
    The layer of areas can be an input polygon layer or it can be square or hexagonal bins calculated
    when the task is run. The tool first determines which points fall within each specified area.
    After determining this point-in-area spatial relationship, statistics about all points in the
    area are calculated and assigned to the area. The most basic statistic is the count of the
    number of points within the area, but you can get other statistics as well.

    For example, suppose you have point features of coffee shop locations and area features of counties,
    and you want to summarize coffee sales by county. Assuming the coffee shops have a TOTAL_SALES attribute,
    you can get the sum of all TOTAL_SALES within each county, the minimum or maximum TOTAL_SALES within each
    county, or other statistics like the count, range, standard deviation, and variance.

    This tool can also work on data that is time-enabled. If time is enabled on the input points, then
    the time slicing options are available. Time slicing allows you to calculate the point-in area relationship
    while looking at a specific slice in time. For example, you could look at hourly intervals, which would
    result in outputs for each hour.

    For an example with time, suppose you had point features of every transaction made at a coffee shop location and no area layer.
    The data has been recorded over a year, and each transaction has a location and a time stamp. Assuming each transaction has a
    TOTAL_SALES attribute, you can get the sum of all TOTAL SALES within the space and time of interest. If these transactions are
    for a single city, we could generate areas that are one kilometer grids, and look at weekly time slices to summarize the
    transactions in both time and space.

    =================================================     ========================================================================
    **Argument**                                          **Description**
    -------------------------------------------------     ------------------------------------------------------------------------
    point_layer                                           Required point feature layer. The point features that will be aggregated
                                                          into the polygons in the ``polygon_layer`` or bins of the specified ``bin_size``.
                                                          See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_type                                              Optional string. If ``polygon_layer`` is not defined, it is required.

                                                          The type of bin that will be generated and into which points will be aggregated.

                                                          Choice list:['Square', 'Hexagon'].

                                                          The default value is "Square".

                                                          When generating bins for Square, the number and units specified determine the height
                                                          and length of the square. For Hexagon, the number and units specified determine the
                                                          distance between parallel sides. Either ``bin_type`` or ``polygon_layer`` must be specified.
                                                          If ``bin_type`` is chosen, ``bin_size`` and ``bin_size_unit`` specifying the size of the bins must be included.
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_size (Required if ``bin_type`` is used)           Optional float. The distance for the bins of type binType that
                                                          the ``point_layer`` will be aggregated into. When generating bins, for Square,
                                                          the number and units specified determine the height and length of the square.
                                                          For Hexagon, the number and units specified determine the distance between parallel sides.
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_size_unit (Required if ``bin_size`` is used)      Optional string. The distance unit for the bins that the ``point_layer`` will be aggregated into.

                                                          Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

                                                          When generating bins for Square, the number and units specified determine the height and
                                                          length of the square. For Hexagon, the number and units specified determine the distance
                                                          between parallel sides. Either ``bin_type`` or ``polygon_layer`` must be specified.
                                                          If ``bin_type`` is chosen, ``bin_size`` and ``bin_size_unit`` specifying the size of the bins must be included.
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_layer                                         Optional polygon feature layer. The polygon features (areas) into which the input points will be aggregated.
                                                          See :ref:`Feature Input<FeatureInput>`.

                                                          One of ``polygon_layer`` or bins ``bin_size`` and  ``bin_size_unit`` is required.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_interval                                    Optional integer. A numeric value that specifies duration of the time step interval. This option is only
                                                          available if the input points are time-enabled and represent an instant in time.

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_interval_unit                               Optional string. A string that specifies units of the time step interval. This option is only available if the
                                                          input points are time-enabled and represent an instant in time.

                                                          Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_repeat_interval                             Optional integer. A numeric value that specifies how often the time step repeat occurs.
                                                          This option is only available if the input points are time-enabled and of time type instant.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_repeat_interval_unit                        Optional string. A string that specifies the temporal unit of the step repeat.
                                                          This option is only available if the input points are time-enabled and of time type instant.

                                                          Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_reference                                   Optional datetime. A date that specifies the reference time to align the time slices to, represented in milliseconds from epoch.
                                                          The default is January 1, 1970, at 12:00 a.m. (epoch time stamp 0). This option is only available if the
                                                          input points are time-enabled and of time type instant.
    -------------------------------------------------     ------------------------------------------------------------------------
    summary_fields                                        Optional list of dicts. A list of field names and statistical summary types that you want to calculate
                                                          for all points within each polygon or bin. Note that the count of points within each polygon is always
                                                          returned. By default, all statistics are returned.

                                                          Example: [{"statisticType": "Count", "onStatisticField": "fieldName1"}, {"statisticType": "Any", "onStatisticField": "fieldName2"}]

                                                          fieldName is the name of the fields in the input point layer.

                                                          statisticType is one of the following for numeric fields:

                                                              * ``Count`` -Totals the number of values of all the points in each polygon.
                                                              * ``Sum`` -Adds the total value of all the points in each polygon.
                                                              * ``Mean`` -Calculates the average of all the points in each polygon.
                                                              * ``Min`` -Finds the smallest value of all the points in each polygon.
                                                              * ``Max`` -Finds the largest value of all the points in each polygon.
                                                              * ``Range`` -Finds the difference between the Min and Max values.
                                                              * ``Stddev`` -Finds the standard deviation of all the points in each polygon.
                                                              * ``Var`` -Finds the variance of all the points in each polygon.

                                                          statisticType is one of the following for string fields:

                                                              * ``Count`` -Totals the number of strings for all the points in each polygon.
                                                              * ``Any` `-Returns a sample string of a point in each polygon.
    -------------------------------------------------     ------------------------------------------------------------------------
    output_name                                           Optional string. The method will create a feature service of the results. You define the name of the service.
    -------------------------------------------------     ------------------------------------------------------------------------
    gis                                                   Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------------------     ------------------------------------------------------------------------
    context                                               Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                              *  Extent (``extent``) - a bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                              * Processing spatial reference (``processSR``) The features will be projected into this coordinate system for analysis.
                                                              * Output Spatial Reference (``outSR``) - the features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                              * Data store (``dataStore``) Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    -------------------------------------------------     ------------------------------------------------------------------------
    future                                                optional Boolean. If True, a GPJob is returned instead of
                                                          results. The GPJob can be queried on the status of the execution.
    =================================================     ========================================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To aggregate number of 911 calls within 1 km summarized by Day count.

            agg_result = aggregate_points(calls,
                                          bin_size=1,
                                          bin_size_unit='Kilometers',
                                          time_step_interval=1,
                                          time_step_interval_unit="Years",
                                          summary_fields=[{"statisticType": "Count", "onStatisticField": "Day"}],
                                          output_name='testaggregatepoints01')
    """

    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Aggregate Points Analysis_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Aggregate Points')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})
    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)
    _set_context(params)

    param_db = {
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "polygon_layer": (_FeatureSet, "polygonLayer"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_repeat_interval": (int, "timeStepRepeatInterval"),
        "time_step_repeat_interval_unit": (str, "timeStepRepeatIntervalUnit"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "AggregatePoints", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

aggregate_points.__annotations__ = {
                     'bin_type': str,
                     'bin_size': float,
                     'bin_size_unit': str,
                     'time_step_interval': int,
                     'time_step_interval_unit': str,
                     'time_step_repeat_interval': int,
                     'time_step_repeat_interval_unit': str,
                     'time_step_reference': _datetime,
                     'summary_fields': str,
                     'output_name': str
                }

def describe_dataset(input_layer,
                     extent_output=False,
                     sample_size=None,
                     output_name=None,
                     gis=None,
                     context=None,
                     future=False,
                     return_tuple=False):
    """
    .. image:: _static/images/describe_dataset/describe_dataset.png 

    The ``describe_dataset`` task provides an overview of your big data.
    The tool outputs a feature layer representing a sample of your 
    input features or a single polygon feature layer that represents the extent of your 
    input features. You can choose to output one, both, or none.

    For example, imagine you are tasked with completing an analysis workflow on a large 
    volume of data. You want to try the workflow, but it could take hours or days with 
    your full dataset. Instead of using time and resources running the full analysis, 
    first create a sample layer to efficiently test your workflow before running it 
    on the full dataset.

    .. note::
        Only available at ArcGIS Enterprise 10.7 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required feature layer. The table, point, line, or polygon feature 
                      layer that will be described, summarized, and sampled. 
                      See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    extent_output     Optional boolean. The task will output a single rectangle
                      feature representing the extent of the ``input_layer`` if this value
                      is set to 'True'. 
                      
                      The default value is 'False'.
    ----------------  ---------------------------------------------------------------
    sample_size       Optional integer. The task will output a feature layer
                      representing a sample of features from the ``input_layer``. Specify
                      the number of sample features to return. If the input value is
                      0 or empty then no sample layer will be created. The output
                      will have the same schema, geometry, and time type as the input
                      layer.
    ----------------  ---------------------------------------------------------------
    output_name       Optional string. The task will create a feature layer of the results. 
                      You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               Optional GIS. The GIS object where the analysis will take place.
    ----------------  ---------------------------------------------------------------
    context           Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                      #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                      #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                      #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                      #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ----------------  ---------------------------------------------------------------
    future            Optional boolean. If True, a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.
    ================  ===============================================================

    :returns: a named tuple with the following keys if ``return_tuple`` is set to 'True':

      "output_json" : dict

      "output" : Table	

      "extent_layer" : feature layer 

      "sample_layer" : feature layer       

      "process_info" : list

    else returns a feature layer of the results.
    
    .. code-block:: python

            # Usage Example: To get an overview of your big data item

            data = describe_dataset(input_layer=big_data_layer, 
                                    extent_output=True,
                                    sample_size=2000,
                                    output_name="describe dataset")
    """
    kwargs = locals()
    tool_name = "DescribeDataset"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Describe_Dataset_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "extent_output" : (bool, "extentOutput"),
        "sample_size" : (int, "sampleSize"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "return_tuple": (bool, "returnTuple"),
        "output_json": (dict, "outputJSON"),
        "output": (_Table, "output"),
        "extent_layer": (_FeatureSet, "extentLayer"),
        "sample_layer": (_FeatureSet, "sampleLayer"),
        "process_info": (list, "processInfo"),
    }

    return_values = [
        {"name": "output_json", "display_name": "Output Dictionary", "type": dict},
        {"name": "output", "display_name": "Output Features", "type": _Table},
        {"name": "extent_layer", "display_name": "Eextent Layer", "type": _FeatureSet},
        {"name": "sample_layer", "display_name": "Sample Layer", "type": _FeatureSet},
        {"name": "process_info", "display_name": "process_info", "type": list},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        res = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        
        if return_tuple:
            return res
        else:
            return output_service        
    except:
        output_service.delete()
        raise 
    return


def join_features(target_layer,
                  join_layer,
                  join_operation="JoinOneToOne",
                  join_fields=None,
                  summary_fields=None,
                  spatial_relationship="Equals",
                  spatial_near_distance=None,
                  spatial_near_distance_unit=None,
                  temporal_relationship=None,
                  temporal_near_distance=None,
                  temporal_near_distance_unit=None,
                  attribute_relationship=None,
                  join_condition=None,
                  output_name=None,
                  gis=None,
                  context=None,
                  future=False):
    """
    .. image:: _static/images/join_features_geo/join_features_geo.png 

    Using either feature layers or tabular data, you can join features and records based on
    specific relationships between the input layers or tables. Joins will be determined by
    spatial, temporal, and attribute relationships, and summary statistics can be optionally
    calculated.

    For example

        * Given point locations of crime incidents with a time, join the crime data to itself
          specifying a spatial relationship of crimes within 1 kilometer of each other and that
          occurred within 1 hour of each other to determine if there are a sequence of crimes
          close to each other in space and time.

        * Given a table of ZIP Codes with demographic information and area features representing
          residential buildings, join the demographic information to the residences so each
          residence now has the information.

    The ``join_features`` task works with two layers. ``join_features`` joins attributes from one
    feature to another based on spatial, temporal, and attribute relationships or some
    combination of the three. The tool determines all input features that meet the specified
    join conditions and joins the second input layer to the first. You can optionally join
    all features to the matching features or summarize the matching features.

    ``join_features`` can be applied to points, lines, areas, and tables. A temporal join
    requires that your input data is time-enabled, and a spatial join requires that your
    data has a geometry.


    ==========================================================================================================  =============================================================================================
    **Argument**                                                                                                **Description**
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    target_layer                                                                                                Required layer. The table, point, line, or polygon features to be joined to. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    join_layer                                                                                                  Required layer. The point, line, or polygon features that will be joined to the ``target_layer``. 
                                                                                                                See :ref:`Feature Input<FeatureInput>`.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    join_operation                                                                                              Optional string. A string representing the type of join that will be applied. 
                                      
                                                                                                                Choice list: ['JoinOneToOne', 'JoinOneToMany'].

                                                                                                                    * ``JoinOneToOne`` - If multiple join features are found that have the same relationships 
                                                                                                                      with a single target feature, the attributes from the multiple join features will be 
                                                                                                                      aggregated using the specified summary statistics. For example, if a point target 
                                                                                                                      feature is found within two separate polygon join features, the attributes from 
                                                                                                                      the two polygons will be aggregated before being transferred to the output point 
                                                                                                                      feature class. If one polygon has an attribute value of 3 and the other has a value 
                                                                                                                      of 7, and a SummaryField of sum is selected, the aggregated value in the output 
                                                                                                                      feature class will be 10. There will always be a Count field calculated, with a 
                                                                                                                      value of 2, for the number of features specified. This is the default.
                                                                                                                    * ``JoinOneToMany`` - If multiple join features are found that have the same relationship 
                                                                                                                      with a single target feature, the output feature class will contain multiple copies (records) 
                                                                                                                      of the target feature. For example, if a single point target feature is found within two 
                                                                                                                      separate polygon join features, the output feature class will contain two copies of the 
                                                                                                                      target feature: one record with the attributes of the first polygon, and another record 
                                                                                                                      with the attributes of the second polygon. There are no summary statistics calculated 
                                                                                                                      with this method.

                                                                                                                The default value is 'JoinOneToOne'.      
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    join_fields                                                                                                 Optional list of dicts. A list of modifications to field names in the joinLayer to be 
                                                                                                                made before completing analysis. Any field that is removed will not have 
                                                                                                                statistics calculated on it.

                                                                                                                Syntax: [{ "action" : "action", "field" : "fieldname1"}, { "action" : "action", "field" : "initial_fieldname", "to" : "new_fieldname"}]

                                                                                                                ``action`` can be the following:

                                                                                                                    * remove - Remove a field from analysis and output .
                                                                                                                    * rename - Rename a field before running analysis.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    summary_fields                                                                                              Optional list of dicts. A list of field names and statistical summary types you want to calculate. 
                                                                                                                Note that the count is always returned. By default, all statistics are returned.

                                                                                                                Syntax: [{"statisticType" : "<statistic type>", "onStatisticField" : "<field name>" }, ...]

                                                                                                                fieldName is the name of the fields in the target layer.

                                                                                                                statisticType is one of the following for numeric fields:

                                                                                                                    * ``Count`` - Totals the number of values of all the points in each polygon.
                                                                                                                    * ``Sum`` - Adds the total value of all the points in each polygon.
                                                                                                                    * ``Mean`` - Calculates the average of all the points in each polygon.
                                                                                                                    * ``Min`` - Finds the smallest value of all the points in each polygon.
                                                                                                                    * ``Max`` - Finds the largest value of all the points in each polygon.
                                                                                                                    * ``Range`` - Finds the difference between the Min and Max values.
                                                                                                                    * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                                                                                                                    * ``Var`` - Finds the variance of all the points in each polygon.

                                                                                                                statisticType is one of the following for string fields:

                                                                                                                    * ``Count`` - Totals the number of strings for all the points in each polygon.
                                                                                                                    * ``Any`` - Returns a sample string of a point in each polygon.  
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    spatial_relationship                                                                                        Optional string. Defines the spatial relationship used to spatially join features.
    
                                                                                                                Choice list: ['Equals', 'Intersects', 'Contains', 'Within', 'Crosses', 'Touches', 'Overlaps', 'Near', 'NearGeodesic']
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    spatial_near_distance (Required if ``spatial_relationship`` is Near or NearGeodesic)                        Optional float.  A float value used for the search distance to determine if 
                                                                                                                the target features are near the join features. This is only applied if ``Near``, 
                                                                                                                or ``NearGeodesic`` is the selected ``spatial_relationship``. You can only enter a 
                                                                                                                single distance value. The units of the distance values are supplied by the 
                                                                                                                ``spatial_near_distance`` parameter.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    spatial_near_distance_unit (Required if ``spatial_relationship`` is Near or NearGeodesic)                   Optional string. The linear unit to be used with the distance value specified in ``spatial_near_distance``.
                                                                                            
                                                                                                                Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    temporal_relationship                                                                                       Optional string. Defines the temporal relationship used to temporally join features.
                                                                                            
                                                                                                                Choice list : ['Equals', 'Intersects', 'During', 'Contains', 
                                                                                                                'Finishes', 'FinishedBy', 'Meets', 'MetBy', 'Overlaps', 'OverlappedBy', 'Starts', 'StartedBy', 'Near']
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    temporal_near_distance (Required if ``temporal_relationship`` is Near, NearBefore, or NearAfter)            Optional integer. An integer value used for the temporal search distance to determine 
                                                                                                                if the target features are temporally near the join features. This is only applied if Near, 
                                                                                                                NearBefore, or NearAfter is the selected ``temporal_relationship``. You can only enter a single 
                                                                                                                distance value. The units of the distance values are supplied by the ``temporal_near_distance_unit`` parameter.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    temporal_near_distance_unit (Required if ``temporal_relationship`` is Near, NearBefore, or NearAfter)       Optional string.The temporal unit to be used with the distance value specified in ``temporal_near_distance``.

                                                                                                                Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    attribute_relationship                                                                                      Optional list of dicts. A target field, relationship, and join field used to join equal attributes. 

                                                                                                                Syntax: [{ "targetField" : "fieldname1", "joinField" : "fieldname2", "operator" : "operator" }]
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    join_condition                                                                                              Optional string. Applies a condition to specified fields. Only features with fields that meet 
                                                                                                                these conditions will be joined. For example, to apply a join to a dataset for only those features 
                                                                                                                where health_spending is greater than 20 percent of income, apply a join condition of target['health_spending'] > (join['income'] * .20) 
                                                                                                                using the field health_spending from the first dataset (``target_layer``) and the income field from the second dataset (``join_layer``).
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    output_name                                                                                                 Optional string. The task will create a feature service of the
                                                                                                                results. You define the name of the service.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    gis                                                                                                         Optional GIS. The GIS object where the analysis will take place.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    context                                                                                                     Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                                                                                #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                                                #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                                                                                #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                                                                #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ----------------------------------------------------------------------------------------------------------  ---------------------------------------------------------------------------------------------
    future                                                                                                      Optional boolean. If 'True', a GPJob is returned instead of
                                                                                                                results. The GPJob can be queried on the status of the execution.

                                                                                                                The default value is 'False'.
    ==========================================================================================================  =============================================================================================

    :Returns: Output Features as Feature Layer Collection Item

    .. code-block:: python

            # Usage Example: To find power outages in your state that may have been caused by a lightning strike. 

            output = join_features(target_layer=outages_layer, 
                                   join_layer=lightning, 
                                   join_operation="JoinOneToMany", 
                                   spatial_relationship="Near", 
                                   spatial_near_distance=20,  
                                   spatial_near_distance_unit="Miles", 
                                   temporal_relationship="NearAfter", 
                                   temporal_near_distance=30, 
                                   temporal_near_distance_unit="Minutes", 
                                   output_name="LightningOutages")
    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Join Features Analysis_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Join Features')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "target_layer": (_FeatureSet, "targetLayer"),
        "join_layer": (_FeatureSet, "joinLayer"),
        "join_operation": (str, "joinOperation"),
        "join_fields": (str, "joinFields"),
        "summary_fields": (str, "summaryFields"),
        "spatial_relationship": (str, "spatialRelationship"),
        "spatial_near_distance": (float, "spatialNearDistance"),
        "spatial_near_distance_unit": (str, "spatialNearDistanceUnit"),
        "temporal_relationship": (str, "temporalRelationship"),
        "temporal_near_distance": (int, "temporalNearDistance"),
        "temporal_near_distance_unit": (str, "temporalNearDistanceUnit"),
        "attribute_relationship": (str, "attributeRelationship"),
        "join_condition": (str, "joinCondition"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},

    ]
    
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "JoinFeatures", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "JoinFeatures", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

join_features.__annotations__ = {
                  'join_operation': str,
                  'join_fields': str,
                  'summary_fields': str,
                  'spatial_relationship': str,
                  'spatial_near_distance': float,
                  'spatial_near_distance_unit': str,
                  'temporal_relationship': str,
                  'temporal_near_distance': int,
                  'temporal_near_distance_unit': str,
                  'attribute_relationship': str,
                  'join_condition': str,
                  'output_name': str}

def reconstruct_tracks(input_layer,
                       track_fields,
                       method="Planar",
                       buffer_field=None,
                       summary_fields=None,
                       distance_split=None,
                       distance_split_unit=None,
                       time_boundary_split=None,
                       time_boundary_split_unit=None,
                       time_boundary_reference=None,
                       output_name=None,
                       gis=None,
                       time_split=None,
                       time_split_unit=None,
                       context=None,
                       future=False):
    """
    .. image:: _static/images/reconstruct_tracks/reconstruct_tracks.png 

    The ``reconstruct_tracks`` task works with a time-enabled layer of either point or polygon 
    features that represents an instant in time. It first determines which features belong to a 
    track using an identifier. Using the time at each location, the tracks are ordered sequentially 
    and transformed into a line or polygon representing the path of movement over time. Optionally, 
    the input can be buffered by a field, which will create a polygon at each location. These buffered 
    points, or polygons if the inputs are polygons, are then joined sequentially to create a track as a 
    polygon where the width is representative of the attribute of interest. Resulting tracks have start 
    and end times that represent the time at the first and last feature in a given track. When the tracks 
    are created, statistics about the input features are calculated and assigned to the output track. The 
    most basic statistic is the count of points within the area, but other statistics can be calculated as 
    well. Features in time-enabled layers can be represented in one of two ways:

        * Instant - A single moment in time
        * Interval - A start and end time
        
    For example, suppose you have GPS measurements of hurricanes every 10 minutes. Each GPS measurement records 
    the hurricane name, location, time of recording, and the wind speed. You could create tracks of the hurricanes 
    using the name of the hurricane as the track identification, and all hurricanes' tracks would be generated. 
    You could calculate statistics such as the mean, maximum, and minimum wind speed of each hurricane, as well 
    as the count of measurements in each track.

    ======================================================================================  ===============================================================
    **Argument**                                                                            **Description**
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    input_layer                                                                             Required layer. The point or polygon features from which tracks 
                                                                                            will be constructed. See :ref:`Feature Input<FeatureInput>`.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    track_fields                                                                            Required string. The fields used to identify distinct tracks. There can 
                                                                                            be multiple ``track_fields``.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    method                                                                                  Optional string. The method used to apply reconstruct tracks and, optionally, 
                                                                                            to apply the buffer. There are two methods to choose from:

                                                                                            * ``Planar`` - This method joins points using a plane method and will not 
                                                                                              cross the international date line. For buffers, this method applies a Euclidean 
                                                                                              buffer and is appropriate for local analysis on projected data. This is the default.
                                                                                            * ``Geodesic`` - This method joins points geodesically and will allow tracks to cross 
                                                                                              the international date line. For buffers, this method is appropriate for large areas 
                                                                                              and any geographic coordinate system.

                                                                                            The default value is 'Planar'.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    buffer_field                                                                            Optional string. A field in the ``input_layer`` that contains a buffer distance or a buffer expression.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    summary_fields                                                                          Optional list of dicts. A list of field names and statistical summary types that you want to calculate for 
                                                                                            all reconstructed track features. By default, all statistics are returned.

                                                                                            Syntax: [{"statisticType" : "<statistic type>", "onStatisticField" : "<field name>" }, ...]

                                                                                            fieldName is the name of the fields in the input point layer.

                                                                                            statisticType is one of the following for numeric fields:

                                                                                                * ``Count`` - Totals the number of values of all the points in each polygon.
                                                                                                * ``Sum`` - Adds the total value of all the points in each polygon.
                                                                                                * ``Mean`` - Calculates the average of all the points in each polygon.
                                                                                                * ``Min`` - Finds the smallest value of all the points in each polygon.
                                                                                                * ``Max`` - Finds the largest value of all the points in each polygon.
                                                                                                * ``Range`` - Finds the difference between the Min and Max values.
                                                                                                * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                                                                                                * ``Var`` - Finds the variance of all the points in each polygon.

                                                                                            statisticType is one of the following for string fields:

                                                                                                * ``Count`` - Totals the number of strings for all the points in each polygon.
                                                                                                * ``Any`` - Returns a sample string of a point in each polygon.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_boundary_split                                                                     Optional integer. A time boundary allows your to analyze values within a defined 
                                                                                            time span. For example, if you use a time boundary of 1 day, starting on January 1st, 
                                                                                            1980 tracks will be analyzed 1 day at a time. The time boundary parameter was introduced 
                                                                                            in ArcGIS Enterprise 10.7.

                                                                                            The ``time_boundary_split`` parameter defines the scale of the time boundary. In the 
                                                                                            case above, this would be 1.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_boundary_split_unit (Required if ``time_boundary_split`` is specified)             Optional string. The unit applied to the time boundary. ``time_boundary_split`` is required 
                                                                                            if a ``time_boundary_split`` is provided.

                                                                                            Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_boundary_reference                                                                 Optional datetime.datetime. A date that specifies the reference time to align the time boundary to, 
                                                                                            represented in milliseconds from epoch.
                                                                                            This option is only available if the ``time_boundary_split`` and ``time_boundary_split_unit`` are set.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    distance_split                                                                          Optional float. A distance used to split tracks. Any features in the ``input_layer`` that are in the same 
                                                                                            track and are greater than this distance apart will be split into a new track. The units of the distance 
                                                                                            values are supplied by the ``distance_split_unit`` parameter.

                                                                                            If both ``distance_split`` and ``time_split`` are used, the track is split when at least one condition is met.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    distance_split_unit (Required if ``distance_split`` is specified)                       Optional string. The distance unit to be used with the distance value specified in ``distance_split``.

                                                                                            Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'NauticalMiles', 'Yards'].
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    output_name                                                                             Optional string. The task will create a feature service of the results. You define the name of the service.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    gis                                                                                     Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_split                                                                              Optional integer. A time duration used to split tracks. Any features in the ``input_layer`` that are in 
                                                                                            the same track and are greater than this time apart will be split into a new track. The units of the distance 
                                                                                            values are supplied by the ``time_split`` parameter.

                                                                                            If both ``distance_split`` and ``time_split`` are used, a track is split when at least one condition is met.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_split_unit  (Required if ``time_split`` is specified)                              Optional string. The temporal unit to be used with the temporal distance value specified in ``time_split``.

                                                                                            Choice list: ['Milliseconds', 'Seconds', 'Minutes', 'Hours', 'Days', 'Weeks', 'Months', 'Years']
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    context                                                                                 Optional dict. The context parameter contains additional settings that affect task execution. 
                                                                                            For this task, there are four settings:

                                                                                            #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                            #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                                                            #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                                            #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    --------------------------------------------------------------------------------------  ---------------------------------------------------------------
    future                                                                                  Optional boolean. If 'True', a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.

                                                                                            The default value is 'False'.                            
    ======================================================================================  ===============================================================

    :returns: feature layer collection item

    .. code-block:: python

            # Usage Example: To reconstruct hurricane tracks.    

            tracks = reconstruct_tracks(input_layer=hurricane_lyr,
                                        track_fields='season, trackID',
                                        method='Geodesic',
                                        buffer_field='size',
                                        summary_fields=[{"statisticType" : "Range", "onStatisticField" : "Wind" }],
                                        distance_split=1,
                                        distance_split_unit='Kilometers',
                                        time_boundary_split=1,
                                        time_boundary_split_unit='Days',
                                        output_name='reconstruct hurricane tracks')
    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Reconstructed Tracks_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Reconstruct Tracks')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)
    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "track_fields": (str, "trackFields"),
        "method": (str, "method"),
        "buffer_field": (str, "bufferField"),
        "summary_fields": (str, "summaryFields"),
        "time_boundary_split" : (int, "timeBoundarySplit"),
        "time_boundary_split_unit" : (str, "timeBoundarySplitUnit"),
        "time_boundary_reference" : (datetime.datetime, "timeBoundaryReference"),
        "distance_split": (int, "distanceSplit"),
        "distance_split_unit": (str, "distanceSplitUnit"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "ReconstructTracks", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "ReconstructTracks", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise


reconstruct_tracks.__annotations__ = {
                       'track_fields':str,
                       'method': str,
                       'buffer_field': str,
                       'summary_fields': str,
                       'time_split': int,
                       'time_split_unit': str,
                       'output_name': str}

def summarize_attributes(input_layer,
                         fields=None,
                         summary_fields=None,
                         output_name=None,
                         gis=None,
                         context=None,
                         future=False):
    """
    .. image:: _static/images/summarize_attributes/summarize_attributes.png 

    ===========================================================================  ===============================================================
    **Argument**                                                                 **Description**
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    input_layer                                                                  Required layer. The features that will be summarized. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    fields                                                                       Optional string. The fields that will be used to summarize like features. For example, 
                                                                                 if you chose a field called property type with the values of commercial and residential, 
                                                                                 all of the features with property type residential would be summarized together with 
                                                                                 summary statistics calculated and all of the commercial features would be summarized together.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    summary_fields                                                               Optional list of dicts. A list of field names and statistical summary types you want 
                                                                                 to calculate for features that are summarized together. Note that the count of features 
                                                                                 with the same fields values is always returned. By default, all statistics are returned.

                                                                                 Syntax: [{"statisticType" : "<statistic type>", "onStatisticField" : "<field name>" }, ...]

                                                                                 fieldName is the name of the fields in the input point layer.

                                                                                 statisticType is one of the following for numeric fields:

                                                                                    * Count - Totals the number of values of all the points in each polygon.
                                                                                    * Sum - Adds the total value of all the points in each polygon.
                                                                                    * Mean - Calculates the average of all the points in each polygon.
                                                                                    * Min - Finds the smallest value of all the points in each polygon.
                                                                                    * Max - Finds the largest value of all the points in each polygon.
                                                                                    * Range - Finds the difference between the Min and Max values.
                                                                                    * Stddev - Finds the standard deviation of all the points in each polygon.
                                                                                    * Var - Finds the variance of all the points in each polygon.

                                                                                 statisticType is one of the following for string fields:

                                                                                    * Count - Totals the number of strings for all the points in each polygon.
                                                                                    * Any - Returns a sample string of a point in each polygon.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    output_name                                                                  Optional string. The task will create a feature service of the results. You define the name of the service.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    gis                                                                          Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    context                                                                      Optional dict. Context contains additional settings that affect task execution. For this task, 
                                                                                 there is one setting:

                                                                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    future                                                                       Optional boolean. If 'True', a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.

                                                                                 The default value is 'False'.
    ===========================================================================  ===============================================================

    :returns: feature layer collection

    .. code-block:: python

            # Usage Example: To summarize similar types of storms to find the sum of property damage.

            summarized_result = summarize_attributes(input_layer=storms, 
                                                     fields="Storm_type",
                                                     summary_fields=[{"statisticType" : "Sum", "onStatisticField" : "PropertyDamage"}],
                                                     output_name="summarized_storms")
    """
    kwargs = locals()


    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Summarize Attributes_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Summarize Attributes')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "fields": (str, "fields"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "SummarizeAttributes", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "SummarizeAttributes", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

summarize_attributes.__annotations__ = {
                         'fields': str,
                         'summary_fields': str,
                         'output_name': str}

def summarize_within(summarized_layer,
                     summary_polygons=None,
                     bin_type=None,
                     bin_size=None,
                     bin_size_unit=None,
                     standard_summary_fields=None,
                     weighted_summary_fields=None,
                     sum_shape=True,
                     shape_units=None,
                     group_by_field=None,
                     minority_majority=False,
                     percent_shape=False,
                     output_name=None,
                     gis=None,
                     context=None,
                     future=False):
    """
    .. image:: _static/images/summarize_within_geo/summarize_within_geo.png 

    The ``summarize_within`` task finds features (and portions of features) that are within the 
    boundaries of areas in the first input layer. The following are examples:

        * Given a layer of watershed boundaries and a layer of land-use boundaries, calculate the total acreage of land-use type for each watershed.
        * Given a layer of parcels in a county and a layer of city boundaries, summarize the average value of vacant parcels within each city boundary.
        * Given a layer of counties and a layer of roads, summarize the total mileage of roads by road type within each county.
  
    You can think of ``summarize_within`` as taking two layers and stacking them on top of each other. One of the layers, 
    ``summary_polygons``, must be a polygon layer, and imagine that these polygon boundaries are all colored red. The other layer, 
    ``summarized_layer``, can be any feature type—point, line, or polygon. After stacking these layers on top of each other, you 
    peer down through the stack and count the number of features in ``summarized_layer`` that fall within the polygons with the 
    red boundaries (``summary_polygons``). Not only can you count the number of features, you can calculate simple statistics about 
    the attributes of the features in ``summarized_layer``, such as sum, mean, minimum, maximum, and so on.

    ===========================================================================  ===============================================================
    **Argument**                                                                 **Description**
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    summarized_layer                                                             Required layer. Point, line, or polygon features that will be summarized for each 
                                                                                 polygon in ``summary_polygons`` or bins. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    summary_polygons                                                             Optional layer. The polygon features. Features, or portions of features, 
                                                                                 in ``summarized_layer`` that fall within the boundaries of these polygons 
                                                                                 will be summarized. You can choose to summarize within a polygon layer that you 
                                                                                 provide or within square or hexagon bins that are generated when the tool runs.
                                                                                 See :ref:`Feature Input<FeatureInput>`.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    bin_type (Required if ``summary_polygons`` is not specified)                 Optional string. The type of bin that will be generated and ``summarized_layer`` will be summarized into. 

                                                                                 Choice list: ['Hexagon', 'Square']

                                                                                 Either ``bin_type`` or ``summary_polygons`` must be specified. If ``bin_type`` is chosen, 
                                                                                 then ``bin_size`` and ``bin_size_unit`` - specifying the size of the bin - must be included.

                                                                                 .. note::
                                                                                       Analysis using Square or Hexagon bins requires a projected coordinate system. 
                                                                                       When aggregating layers into bins, the input layer or processing extent (``processSR``) 
                                                                                       must have a projected coordinate system. At 10.5.1, 10.6, and 10.6.1, if a projected 
                                                                                       coordinate system is not specified when running analysis, the World Cylindrical Equal 
                                                                                       Area (WKID 54034) projection will be used. At 10.7 or later, if a projected coordinate 
                                                                                       system is not specified when running analysis, a projection will be picked based on 
                                                                                       the extent of the data.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    bin_size  (Required if ``bin_type`` is specified)                            Optional float. The distance for the bins of type ``bin_type``. 
                                                                                 When generating bins, for Square, the number and units specified determine the 
                                                                                 height and length of the square, and for Hexagon, the number and units specified 
                                                                                 determine the distance between parallel sides.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    bin_size_unit (Required if ``bin_size`` is specified)                        Optional string. The linear distance unit for the bins that ``summarized_layer`` will be summarized into.
                                                     
                                                                                 Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    standard_summary_fields                                                      Optional list of dicts. A list of field names and statistical summary type that you want to calculate 
                                                                                 for all features in ``summarized_layer`` that are within each polygon in ``summary_polygons`` or bins. 
                                                                                 The standard statistics are calculated using the whole attribute values from any feature 
                                                                                 that is within ``summary_polygons``.

                                                                                 Syntax: [{"statisticType" : "<statistic type>", "onStatisticField" : "<field name>" }, ...]

                                                                                 fieldName is the name of the fields in the input point layer.

                                                                                 statisticType is one of the following for numeric fields:

                                                                                    * ``Count`` - Totals the number of features in each polygon.
                                                                                    * ``Sum`` - Adds the total value of all the features in each polygon.
                                                                                    * ``Mean`` - Calculates the average of all the features in each polygon.
                                                                                    * ``Min`` - Finds the smallest value of all the features in each polygon.
                                                                                    * ``Max`` - Finds the largest value of all the features in each polygon.
                                                                                    * ``Range`` - Finds the difference between Min and Max.
                                                                                    * ``Stddev`` - Finds the standard deviation of all the features in each polygon.
                                                                                    * ``Var`` - Finds the variance of all the features in each polygon.

                                                                                 statisticType is one of the following for string fields:

                                                                                    * ``Count`` - Totals the number of strings for all the features in each polygon.
                                                                                    * ``Any`` - Returns a sample string of a feature in each polygon.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    weighted_summary_fields                                                      Optional list of dicts. A list of field names and statistical summary type that you want to calculate 
                                                                                 for all features in ``summarized_layer`` that are within each polygon in ``summary_polygons`` or bins. 
                                                                                 The weighted statistics are calculated using the geographically weighted attribute values 
                                                                                 from features that are within ``summary_polygons``. Resulting fields from proportional statistics 
                                                                                 will be denoted with a p. Weighted statistics can only be applied to ``summarized_layer`` that 
                                                                                 are lines or polygons.

                                                                                 Syntax: [{"statisticType" : "<statistic type>", "onStatisticField" : "<field name>" }, ...]

                                                                                 fieldName is the name of the fields in the input point layer.

                                                                                 statisticType is one of the following for numeric fields:

                                                                                    * ``Count`` - The count of each field multiplied by the proportion of the summarized layer within the polygons.
                                                                                    * ``Sum`` - The sum of weighted of values in each field. Where the weight applied is the proportion of the summarized layer within the polygons.
                                                                                    * ``Mean`` - The weighted mean of values in each field. Where the weight applied is the proportion of the summarized layer within the polygons.
                                                                                    * ``Min`` - The minimum of weighted values in each field. Where the weight applied is the proportion of the summarized layer within the polygons.
                                                                                    * ``Max`` - The maximum of weighted values in each field. Where the weight applied is the proportion of the summarized layer within the polygons.
                                                                                    * ``Range`` - Finds the difference between Min and Max.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    sum_shape                                                                    Optional boolean. A boolean value that instructs the task to calculate statistics based on the 
                                                                                 shape type of ``summarized_layer``, such as the length of lines or areas of polygons 
                                                                                 of ``summarized_layer`` within each polygon in ``summary_polygons``.

                                                                                 The default value is 'True'.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    shape_units                                                                  Optional string. The units used to calculate ``sum_shape``.

                                                                                 Values:

                                                                                    * When ``summarized_layer`` contains polygons, Choice list: ['Acres', 'Hectares', 'SquareMeters', 'SquareKilometers', 'SquareMiles', 'SquareYards', 'SquareFeet'].
                                                                                    * When ``summarized_layer`` contains lines, Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    group_by_field                                                               Optional string. This is a field of the ``summarized_layer`` features that you can use to calculate 
                                                                                 statistics separately for each unique attribute value. For example, suppose the ``summarized_layer`` contains city boundaries and the summaryPolygons features are parcels. `
                                                                                 One of the fields of the parcels is Status which contains two values: VACANT and OCCUPIED. 
                                                                                 To calculate the total area of vacant and occupied parcels within the boundaries of cities, 
                                                                                 use Status as the ``group_by_field`` field. This parameter is available at ArcGIS Enterprise 10.6.1+.

                                                                                 When a ``group_by_field`` field is provided, the service returns a table containing the 
                                                                                 statistics in the groupBySummaryoutput parameter.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    minority_majority                                                            Optioal boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified. 
                                                                                 If true, the minority (least dominant) or the majority (most dominant) attribute values 
                                                                                 for each group field are calculated. Two new fields are added to the ``result_layer`` prefixed with 
                                                                                 Majority_ and Minority_. This parameter is available at ArcGIS Enterprise 10.6.1+.

                                                                                 The default value is 'False'.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    percent_shape                                                                Optioal boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified. 
                                                                                 If set to true, the percentage of each unique ``group_by_field`` value is calculated for 
                                                                                 each sum within layer polygon. The default is false. This parameter is available at ArcGIS Enterprise 10.6.1+.

                                                                                 The default value is 'False'.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    output_name                                                                  Optional string. The task will create a feature service of the results. You define the name of the service.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    context                                                                      Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                 #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                                                 #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    gis                                                                          Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ---------------------------------------------------------------------------  ---------------------------------------------------------------
    future                                                                       Optional boolean. If 'True', a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.
    ===========================================================================  ===============================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To calculate the distance and average slope of bike lanes within each city district.

            summarize_within_result = summarize_within(summary_polygons=districts, 
                                                       summarized_layer=bike_lanes,
                                                       weighted_summary_fields=[{"statisticType" : "Average","onStatisticField" : "Slope"}],
                                                       output_name="summary_of_bike_lanes")
    """
    kwargs = locals()


    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Summarize Within_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Summarize Within')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "summary_polygons": (_FeatureSet, "summaryPolygons"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "summarized_layer": (_FeatureSet, "summarizedLayer"),
        "standard_summary_fields": (str, "standardSummaryFields"),
        "weighted_summary_fields": (str, "weightedSummaryFields"),
        "sum_shape": (bool, "sumShape"),
        "shape_units": (str, "shapeUnits"),
        "group_by_field": (str, "groupByField"),
        "minority_majority" : (bool, "minorityMajority"),
        "percent_shape" : (bool, "percentShape"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "SummarizeWithin", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "SummarizeWithin", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

summarize_within.__annotations__ = {
                     'bin_type': str,
                     'bin_size': float,
                     'bin_size_unit': str,
                     'standard_summary_fields': str,
                     'weighted_summary_fields': str,
                     'sum_shape': bool,
                     'shape_units': str,
                     'group_by_field': str,
                     'minority_majority' : bool,
                     'percent_shape' : bool,
                     'output_name': str
                }




import math
from functools import reduce

import pandas as pd
import numpy as np

from arcgis.geometry._types import Geometry
from arcgis.features.geo._accessor import GeoAccessor
from arcgis.features.geo._accessor import GeoSeriesAccessor
from arcgis.features.geo._accessor import _is_geoenabled
from arcgis.features.geo._array import GeoArray
_HASARCPY, _HASSHAPELY = None, None

#----------------------------------------------------------------------
def _check_geometry_engine():
    """checks if the geometry engine exists"""
    global _HASARCPY
    global _HASSHAPELY
    if _HASARCPY is None:
        try:
            import arcpy
            _HASARCPY = True
        except:
            _HASARCPY = False
    if _HASSHAPELY is None:
        try:
            import shapely
            _HASSHAPELY = True
        except:
            _HASSHAPELY = False
    return _HASARCPY, _HASSHAPELY
#--------------------------------------------------------------------------
def _overlay_difference(df1, df2):
    """
    Overlay Difference operation used in overlay function
    """
    # Spatial Index to create intersections
    spatial_index = df2.spatial.sindex()
    bbox = df1[df1.spatial.name].geom.extent
    fn = lambda x: spatial_index.intersect(x)
    sidx = bbox.apply(fn)
    # Create differences
    new_g = []
    for geom, neighbours in zip(df1[df1.spatial.name], sidx):
        new = reduce(lambda x, y: x.difference(y).buffer(0),
                     [geom] + list(df2[df2.spatial.name].iloc[neighbours]))
        new_g.append(new)
    differences = pd.Series(GeoArray(values=new_g, copy=True), index=df1.index)
    q = differences.isnull()
    geom_diff = differences[~q].copy()
    dfdiff = df1[~q].copy()
    dfdiff[dfdiff.spatial.name] = geom_diff
    return dfdiff
#--------------------------------------------------------------------------
def _symmetric_difference(df1, df2):
    """
    The symmetric difference, also known as the disjunctive union, of two
    sets is the set of elements which are in either of the sets and not in
    their intersection.

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    dfdiff1 = _overlay_difference(df1, df2)
    dfdiff2 = _overlay_difference(df2, df1)
    dfdiff1['__idx1'] = range(len(dfdiff1))
    dfdiff2['__idx2'] = range(len(dfdiff2))
    dfdiff1['__idx2'] = np.nan
    dfdiff2['__idx1'] = np.nan
    dfsym = dfdiff1.merge(dfdiff2, on=['__idx1', '__idx2'], how='outer',
                          suffixes=['_1', '_2'])
    geometry = dfsym.SHAPE_1.copy()
    geometry[dfsym.SHAPE_1.isnull()] = \
        dfsym.loc[dfsym.SHAPE_1.isnull(), 'SHAPE_2']
    dfsym.drop(['SHAPE_1', 'SHAPE_2'], axis=1, inplace=True)
    dfsym.reset_index(drop=True, inplace=True)
    dfsym = pd.DataFrame(data=dfsym)
    dfsym['SHAPE'] = geometry
    dfsym.spatial.name
    return dfsym
#--------------------------------------------------------------------------
def _overlay_intersection(df1, df2):
    """
    Overlay Intersection operation used in overlay function
    """
    def _ops_intersection(this, other):
        """performs the series vs series intersection"""
        lu = {
            "polygon" : 4,
            "point" : 1,
            "multipoint" : 1,
            "line" : 2,
            "polyline" : 2

        }
        gtype = lu[this.geom.geometry_type[0]]
        this, other = this.align(other)
        data = np.array(
                [getattr(this_elem, 'intersect')(other_elem, gtype)
                 if not this_elem.is_empty | other_elem.is_empty else null_value
                 for this_elem, other_elem in zip(this, other)])

        return pd.Series(GeoArray(data), index=this.index)
    # Spatial Index to create intersections
    spatial_index = df2.spatial.sindex()
    bbox = df1[df1.spatial.name].geom.extent
    fn = lambda x: spatial_index.intersect(x)
    sidx = bbox.apply(fn)
    # Create pairs of geometries in both dataframes to be intersected
    nei = []
    for i, j in enumerate(sidx):
        for k in j:
            nei.append([i, k])
    if nei != []:
        pairs = pd.DataFrame(nei, columns=['__idx1', '__idx2'])
        left = df1[df1.spatial.name].take(pairs['__idx1'].values)
        left.reset_index(drop=True, inplace=True)
        right = df2[df2.spatial.name].take(pairs['__idx2'].values)
        right.reset_index(drop=True, inplace=True)
        intersections = _ops_intersection(left, right).geom.buffer(0)
        intersections.name = "intersections"

        # only keep actual intersecting geometries
        pairs_intersect = pairs[~intersections.isnull()]
        geom_intersect = intersections[~intersections.isnull()]

        # merge data for intersecting geometries
        df1 = df1.reset_index(drop=True)
        df2 = df2.reset_index(drop=True)
        dfinter = pairs_intersect.merge(
            df1.drop(df1.spatial.name, axis=1),
            left_on='__idx1', right_index=True)
        dfinter = dfinter.merge(
            df2.drop(df2.spatial.name, axis=1),
            left_on='__idx2', right_index=True, suffixes=['_1', '_2'])
        dfinter['SHAPE'] = geom_intersect
        return dfinter
    else:
        return pd.DataFrame(data=[],
                            columns=list(set(df1.columns).union(df2.columns)) + ['__idx1', '__idx2'])
#--------------------------------------------------------------------------
def _erase(df1, df2):
    """
    This overlay operation erases geometries by the second SeDF object.

    """
    # Spatial Index to create intersections
    spatial_index = df2.sindex
    bbox = df1.geometry.apply(lambda x: x.bounds)
    sidx = bbox.apply(lambda x: list(spatial_index.intersection(x)))
    # Create differences
    new_g = []
    for geom, neighbours in zip(df1.geometry, sidx):
        new = reduce(lambda x, y: x.difference(y).buffer(0),
                     [geom] + list(df2.geometry.iloc[neighbours]))
        new_g.append(new)
    differences = GeoSeries(new_g, index=df1.index)
    geom_diff = differences[~differences.is_empty].copy()
    dfdiff = df1[~differences.is_empty].copy()
    dfdiff[dfdiff._geometry_column_name] = geom_diff
    return dfdiff

def _overlay_union(df1, df2):
    """
    Overlay Union operation used in overlay function
    """
    dfinter = _overlay_intersection(df1, df2)
    dfsym = _symmetric_difference(df1, df2)
    dfunion = pd.concat([dfinter, dfsym], ignore_index=True, sort=False)
    columns = list(dfunion.columns)
    return dfunion.reindex(columns=columns)

def overlay(sdf1, sdf2, op="union"):
    """
    Perform spatial overlay operations between two polygons Spatially
    Enabled DataFrames.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf1                         Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    sdf2                         Required Spatially Enabled DataFrame. The geometry to perform the operation from.
    -------------------------    ---------------------------------------------------------
    op                           Optional String. The spatial operation to perform.  The
                                 allowed value are: union, erase, identity, intersection
    =========================    =========================================================

    :returns: Spatially enabled DataFrame (pd.DataFrame)

    """
    allowed_hows = [
        'intersection',
        'union',
        'identity',
        'symmetric_difference',
        'difference',
        'erase'
    ]

    op = str(op).lower()

    if sdf1.spatial.geometry_type != sdf2.spatial.geometry_type:
        raise ValueError(("Spatially enabled DataFrame must "
                          "be the same geometry type."))

    if op == 'symmetric_difference'and \
       sdf1.spatial.geometry_type != ['polygon'] and \
       sdf2.spatial.geometry_type != ['polygon']:
        raise ValueError(("symmetric_difference is only supported for "
                          "polygon geometries."))


    _hasao, _hasshp = _check_geometry_engine()

    if _hasao == False and _hasshp and \
       sdf1.spatial.geometry_type != ['polygon'] and \
       sdf2.spatial.geometry_type != ['polygon']:
        raise ValueError(("Using shapely's geometry engine only "
                          "support Polygon geometries."))


    if (_hasao or _hasshp) and \
       op in allowed_hows:
        if op in ['union', 'identity']:
            return _overlay_union(sdf1, sdf2)
        elif op in ['difference', 'erase']:
            return _overlay_difference(sdf1, sdf2)
        elif op == 'intersection':
            return _overlay_intersection(sdf1, sdf2)
        elif op == 'symmetric_difference'and \
             sdf1.spatial.geometry_type == ['polygon'] and \
             sdf2.spatial.geometry_type == ['polygon']:
            return _symmetric_difference(sdf1, sdf2)
    return None



"""
Allows for dataset to dataset comparisons by passing in DataFrames or
Geometries.
"""
import math
from functools import reduce

import pandas as pd
import numpy as np

from arcgis.geometry._types import Geometry, Point, Polygon, Polyline, MultiPoint
from arcgis.features.geo._accessor import GeoAccessor
from arcgis.features.geo._accessor import GeoSeriesAccessor
from arcgis.features.geo._accessor import _is_geoenabled
from arcgis.features.geo._array import GeoArray

_HASARCPY, _HASSHAPELY = None, None

#----------------------------------------------------------------------
def _check_geometry_engine():
    """checks if the geometry engine exists"""
    global _HASARCPY
    global _HASSHAPELY
    if _HASARCPY is None:
        try:
            import arcpy
            _HASARCPY = True
        except:
            _HASARCPY = False
    if _HASSHAPELY is None:
        try:
            import shapely
            _HASSHAPELY = True
        except:
            _HASSHAPELY = False
    return _HASARCPY, _HASSHAPELY
#--------------------------------------------------------------------------
def contains(sdf, other, relation="CLEMENTINI"):
    """
    Indicates if the base geometry contains the comparison geometry.

    `contains` is the opposite of `within`.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    -------------------------    ---------------------------------------------------------
    relation                     Optional String.  The spatial relationship type.  The
                                 allowed values are: BOUNDARY, CLEMENTINI, and PROPER.

                                 BOUNDARY - Relationship has no restrictions for interiors
                                            or boundaries.
                                 CLEMENTINI - Interiors of geometries must intersect. This
                                              is the default.
                                 PROPER - Boundaries of geometries must not intersect.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.contains(other, relation)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.contains(g, relation)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def crosses(sdf, other):
    """

    Indicates if the two geometries intersect in a geometry of a lesser
    shape type.

    Two polylines cross if they share only points in common, at least one
    of which is not an endpoint. A polyline and an polygon cross if they
    share a polyline or a point (for vertical line) in common on the
    interior of the polygon which is not equivalent to the entire polyline.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.crosses(other)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.crosses(g)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def disjoint(sdf, other):
    """
    Indicates if the base and comparison geometries share no points in common.

    **Two geometries intersect if `disjoint` returns False.**

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([True] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.disjoint(other)
        dj.index = sub.index
        ud = ~(~ud | ~dj)
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.disjoint(g)
                dj.index = sub.index
                ud = ~(~ud | ~dj)
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def equals(sdf, other):
    """

    Indicates if the base and comparison geometries are of the same shape
    type and define the same set of points in the plane. This is a 2D
    comparison only; M and Z values are ignored.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.equals(other)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.equals(g)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def overlaps(sdf, other):
    """

    Indicates if the intersection of the two geometries has the same shape
    type as one of the input geometries and is not equivalent to either of
    the input geometries.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.overlaps(other)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.overlaps(g)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def touches(sdf, other):
    """

    Indicates if the boundaries of the geometries intersect.

    Two geometries touch when the intersection of the geometries is not
    empty, but the intersection of their interiors is empty. For example,
    a point touches a polyline only if the point is coincident with one of
    the polyline end points.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.touches(other)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.touches(g)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None
#--------------------------------------------------------------------------
def within(sdf, other, relation="CLEMENTINI"):
    """

    Indicates if the base geometry is within the comparison geometry.

    `within` is the opposite operator of `contains`.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    -------------------------    ---------------------------------------------------------
    relation                     Optional String.  The spatial relationship type.  The
                                 allowed values are: BOUNDARY, CLEMENTINI, and PROPER.

                                 BOUNDARY - Relationship has no restrictions for interiors
                                            or boundaries.
                                 CLEMENTINI - Interiors of geometries must intersect. This
                                              is the default.
                                 PROPER - Boundaries of geometries must not intersect.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    global _HASARCPY, _HASSHAPELY

    if _HASARCPY == False and _HASSHAPELY == False:
        return None

    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.within(other, relation)
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.within(other, relation)
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None



import os
import datetime
from concurrent.futures import Future
import logging
import json
_log = logging.getLogger(__name__)
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData

class GPJob(object):
    """
    Represents a Single Geoprocessing Job.  The `GPJob` class allows for the asynchronous operation
    of any geoprocessing task.  To request a GPJob task, the code must be called with `future=True`
    or else the operation will occur synchronously.  This class is not intended for users to call
    directly.


    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    future            Required ccurrent.futures.Future.  The async object created by
                      the geoprocessing (GP) task.
    ----------------  ---------------------------------------------------------------
    gptool            Required Layer. The Geoprocessing Service
    ----------------  ---------------------------------------------------------------
    jobid             Required String. The unique ID of the GP Job.
    ----------------  ---------------------------------------------------------------
    task_url          Required String. The URL to the GP Task.
    ----------------  ---------------------------------------------------------------
    gis               Required GIS. The GIS connection object
    ----------------  ---------------------------------------------------------------
    notify            Optional Boolean.  When set to True, a message will inform the
                      user that the geoprocessing task has completed. The default is
                      False.
    ================  ===============================================================

    """
    _future = None
    _jobid = None
    _url = None
    _gis = None
    _task_name = None
    _is_fa = False
    _is_ra = False
    _is_ortho = False
    _start_time = None
    _end_time = None
    _item_properties = None
    #----------------------------------------------------------------------
    def __init__(self, future, gptool, jobid, task_url, gis, notify=False):
        """
        initializer
        """
        assert isinstance(future, Future)
        self._future = future
        self._start_time = datetime.datetime.now()
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)
        self._gptool = gptool
        self._jobid = jobid
        self._url = task_url
        self._gis = gis
    #----------------------------------------------------------------------
    @property
    def ellapse_time(self):
        """
        Returns the Ellapse Time for the Job
        """
        if self._end_time:
            return self._end_time - self._start_time
        else:
            return datetime.datetime.now() - self._start_time
    #----------------------------------------------------------------------
    def _set_end_time(self, future):
        """sets the finish time"""
        self._end_time = datetime.datetime.now()
    #----------------------------------------------------------------------
    def _notify(self, future):
        """prints finished method"""
        jobid = str(self).replace("<", "").replace(">", "")
        try:
            res = future.result()
            infomsg = '{jobid} finished successfully.'.format(jobid=jobid)
            _log.info(infomsg)
            print(infomsg)
        except Exception as e:
            msg = str(e)
            msg = '{jobid} failed: {msg}'.format(jobid=jobid, msg=msg)
            _log.info(msg)
            print(msg)
    #----------------------------------------------------------------------
    def __str__(self):
        return "<%s GP Job: %s>" % (self.task, self._jobid)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<%s GP Job: %s>" % (self.task, self._jobid)
    #----------------------------------------------------------------------
    @property
    def task(self):
        """Returns the task name.
        :returns: string
        """
        if self._task_name is None:
            self._task_name = os.path.basename(self._url)
        return self._task_name
    #----------------------------------------------------------------------
    @property
    def status(self):
        """
        returns the GP status

        :returns: String
        """
        url = self._url + "/jobs/%s" % self._jobid
        params = {'f' : 'json',
                  'returnMessages': True}


        res = self._gis._con.post(url, params)
        if 'jobStatus' in res:
            return res['jobStatus']
        return res
    #----------------------------------------------------------------------
    def cancel(self):
        """
        Attempt to cancel the call. If the call is currently being executed
        or finished running and cannot be cancelled then the method will
        return False, otherwise the call will be cancelled and the method
        will return True.

        :returns: boolean
        """
        if self.done():
            return False
        if self.cancelled():
            return False
        try:
            url = self._url + "/jobs/%s/cancel" % self._jobid
            params = {'f' : 'json'}
            res = self._gis._con.post(url, params)
            if 'jobStatus' in res:
                self._future.set_result({'jobStatus' : 'esriJobCancelled'})
                self._future.cancel()
                return True
            self._future.set_result({'jobStatus' : 'esriJobCancelled'})
            self._future.cancel()
            #self._future.set_result({'jobStatus' : 'esriJobCancelled'})
            return res
        except:
            self._future.cancel()
        return True
    #----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :returns: boolean
        """
        return self._future.cancelled()
    #----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :returns: boolean
        """
        return self._future.running()
    #----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :returns: boolean
        """
        return self._future.done()
    #----------------------------------------------------------------------
    def result(self):
        """
        Return the value returned by the call. If the call hasn't yet completed
        then this method will wait.

        :returns: object
        """
        if self.cancelled():
            return None
        if self._is_fa:
            return self._process_fa(self._future.result())
        elif self._is_ra:
            return self._process_ra(self._future.result())
        elif self._is_ortho:
            return self._process_ortho(self._future.result())
        return self._future.result()

    def _process_ortho(self, result):
        """handles the ortho imagery response"""
        import arcgis

        if hasattr(result, '_fields'):
            r = {}
            iids = []
            for key in result._fields:
                value = getattr(result, key)
                if isinstance(value, dict) and 'featureSet' in value:
                    r[key] = arcgis.features.FeatureCollection(value)
                elif isinstance(value, dict) and 'url' in value and value['url'].lower().find("imageserver"):
                    return arcgis.raster.ImageryLayer(url=value['url'], gis=self._gis)
                elif isinstance(value, dict) and 'url' in value and value['url'].lower().find("featureserver"):
                    return arcgis.features.FeatureLayerCollection(url=value['url'], gis=self._gis)
                elif isinstance(value, dict) and 'itemId' in value and len(value['itemId']) > 0:
                    if not value['itemId'] in iids:
                        r[key] = arcgis.gis.Item(self._gis, value['itemId'])
                        iids.append(value['itemId'])
                elif len(str(value)) > 0 and value:
                    r[key] = value
            if len(r) == 1:
                return r[list(r.keys())[0]]

            return r
        else:
            value = result
            if isinstance(value, (DataFile, RasterData, LinearUnit)):
                return value
            elif isinstance(value, str) and value.lower().find('imageserver') >-1:
                return arcgis.raster.ImageryLayer(url=value, gis=self._gis)
            elif isinstance(value, (dict, tuple, list)) == False:
                return value
            elif 'itemId' in value and \
               len(value['itemId']) > 0:
                itemid = value['itemId']
                return arcgis.gis.Item(gis=self._gis, itemid=itemid)
            elif isinstance(value, dict) and "items" in value:
                itemid = list(value['items'].keys())[0]
                return arcgis.gis.Item(gis=self._gis, itemid=itemid)
            elif self.task == 'QueryCameraInfo':
                import pandas as pd
                columns = value['schema']
                data = value['content']
                return pd.DataFrame(data, columns=columns)
            elif isinstance(value, dict) and 'url' in value and value['url'].lower().find("imageserver"):
                return arcgis.raster.ImageryLayer(url=value['url'], gis=self._gis)
            elif isinstance(value, dict) and 'url' in value and value['url'].lower().find("featureserver"):
                return arcgis.features.FeatureLayerCollection(url=value['url'], gis=self._gis)
            elif isinstance(value, dict) and 'featureSet' in value:
                return arcgis.features.FeatureCollection(value)
            return value
        return result

    def _process_ra(self, result):
        import arcgis
        if isinstance(result, arcgis.features.FeatureLayer):
            if self._item_properties:
                _item_properties = {
                    "properties": {
                        "jobUrl": self._url + '/jobs/' + self._jobid,
                        "jobType": "GPServer",
                        "jobId": self._jobid,
                        "jobStatus": "completed"
                    }
                }
                self._return_item.update(item_properties=_item_properties)
            return self._return_item
        if hasattr(result, '_fields'):
            r = {}
            iids = []
            for key in result._fields:
                value = getattr(result, key)
                if isinstance(value, dict) and 'featureSet' in value:
                    r[key] = arcgis.features.FeatureCollection(value)
                elif isinstance(value, dict) and 'itemId' in value and len(value['itemId']) > 0:
                    if not value['itemId'] in iids:
                        r[key] = arcgis.gis.Item(self._gis, value['itemId'])
                        if self._item_properties:
                            _item_properties = {
                                "properties": {
                                    "jobUrl": self._url + '/jobs/' + self._jobid,
                                    "jobType": "GPServer",
                                    "jobId": self._jobid,
                                    "jobStatus": "completed"
                                }
                            }
                            r[key].update(item_properties=_item_properties)
                        iids.append(value['itemId'])
                elif len(str(value)) > 0 and value:
                    r[key] = value
            if len(r) == 1:
                return r[list(r.keys())[0]]
            if (self.task == "CalculateDistance" or
                self.task == "DetermineOptimumTravelCostNetwork" or
                self.task == "FlowDirection" or
                self.task == "CalculateTravelCost"):
                m = {}
                if isinstance (r,dict):
                    for key, value in r.items():
                        m[key[0:key.rindex('_')+1]+'service'] = r[key]
                    r=m

            if(self.task == 'InterpolatePoints'):
                if "process_info" in  r.keys():
                    process_info = r['process_info']
                    html_final="<b>The following table contains cross validation statistics:</b><br></br><table style='width: 250px;margin-left: 2.5em;'><tbody>"
                    for row in process_info:
                        temp_dict=json.loads(row)
                        if isinstance(temp_dict["message"],list):
                            html_final+="<tr><td>"+temp_dict["message"][0]+"</td><td style='float:right'>"+temp_dict["params"][temp_dict["message"][1].split("${")[1].split("}")[0]]+"</td></tr>"

                    html_final+="</tbody></table><br></br>"
                    from IPython.display import HTML
                    process_info_html = HTML(html_final)
                    r['process_info'] = process_info_html
                    r['output_raster'].update(item_properties={"description":html_final})

            return_value_names = []
            for key, value in r.items():
                return_value_names.append(key)
            num_returns = len(r)
            if num_returns == 1:
                return r[return_value_names[0]]

            else:
                ret_names = []
                for return_value in return_value_names:
                    ret_names.append(return_value)
                import collections
                NamedTuple = collections.namedtuple('FunctionOutput', ret_names)
                function_output = NamedTuple(**r)
                return function_output

        elif isinstance(result, arcgis.raster.ImageryLayer):
            return result
        else:
            value = result
            if isinstance(value, dict) and 'itemId' in value and \
               len(value['itemId']) > 0:
                itemid = value['itemId']
                item = arcgis.gis.Item(gis=self._gis, itemid=itemid)
                if self._item_properties:
                    _item_properties = {
                        "properties": {
                            "jobUrl": self._url + '/jobs/' + self._jobid,
                            "jobType": "GPServer",
                            "jobId": self._jobid,
                            "jobStatus": "completed"
                        }
                    }
                    item.update(item_properties=_item_properties)
                return item
            elif isinstance(value, dict) and 'url' in value:
                return value["url"]
            elif isinstance(value, dict) and 'contentList' in value:
                if value is "":
                    return None
                elif isinstance(value["contentList"], str):
                    return json.loads(value["contentList"])
                return value['contentList']
            elif isinstance(value, dict) and 'modelInfo' in value:
                try:
                    dict_output =  json.loads(value["modelInfo"])
                    return dict_output
                except:
                    return value
            elif isinstance(value, dict) and 'result' in value:
                return value["result"]
            elif isinstance(value, dict) and "items" in value:
                itemid = list(value['items'].keys())[0]
                item = arcgis.gis.Item(gis=self._gis, itemid=itemid)
                if self._item_properties:
                    _item_properties = {
                        "properties": {
                            "jobUrl": self._url + '/jobs/' + self._jobid,
                            "jobType": "GPServer",
                            "jobId": self._jobid,
                            "jobStatus": "completed"
                        }
                    }
                    item.update(item_properties=_item_properties)
                return item
            elif isinstance(value, dict) and 'featureSet' in value:
                return arcgis.features.FeatureCollection(value)
            elif isinstance(value, list) and value is not None:
                output_model_list = []
                from arcgis.learn import Model
                for element in value:
                    if isinstance(element,dict):
                        if "id" in element.keys():
                            item = arcgis.gis.Item(gis=self._gis, itemid=element["id"])
                            output_model_list.append(Model(item))
                return output_model_list

        return result


    def _process_fa(self, result):
        import arcgis
        HAS_ITEM = False
        if hasattr(result, '_fields'):
            r = {}
            iids = []
            for key in result._fields:
                value = getattr(result, key)
                if self.task in ['AggregatePoints', 'ConnectOriginsToDestinations', 
                                 'SummarizeNearby', "InterpolatePoints"] and \
                   isinstance(value, dict) and 'featureSet' in value:
                    r[key] = arcgis.features.FeatureCollection(value)
                elif isinstance(value, dict) and 'featureSet' in value:
                    HAS_ITEM = True
                    r[key] = arcgis.features.FeatureCollection(value)
                elif isinstance(value, dict) and 'itemId' in value and len(value['itemId']) > 0:
                    if not value['itemId'] in iids:
                        HAS_ITEM = True
                        r[key] = arcgis.gis.Item(self._gis, value['itemId'])
                        iids.append(value['itemId'])
                elif len(str(value)) > 0 and value:
                    r[key] = value
                elif HAS_ITEM == False and \
                     (self.task in ['AggregatePoints', 'CreateWatersheds', 'PlanRoutes',
                                   'ConnectOriginsToDestinations',
                                   'SummarizeNearby', "InterpolatePoints"] or \
                      self.task == 'ConnectOriginsToDestinations'):
                    r[key] = value
            if len(r) == 1:
                return r[list(r.keys())[0]]
            return r
        else:
            value = result
            if 'itemId' in value and \
               len(value['itemId']) > 0:
                itemid = value['itemId']
                return arcgis.gis.Item(gis=self._gis, itemid=itemid)
            elif self.task.lower() == 'createroutelayers':
                return [arcgis.gis.Item(gis=self._gis, itemid=itemid) for itemid in result['items']]
            elif isinstance(value, dict) and "items" in value and len(set(value['items'].keys())) == 1:
                itemid = list(value['items'].keys())[0]
                return arcgis.gis.Item(gis=self._gis, itemid=itemid)
            elif isinstance(value, dict) and "items" in value and len(set(value['items'].keys())) > 1:
                return [arcgis.gis.Item(gis=self._gis, itemid=itemid) for itemid in result['items']]
            elif isinstance(value, dict) and 'featureSet' in value:
                return arcgis.features.FeatureCollection(value)
            return value
        return result

